import enum as E
import math
import random
import typing as T

import torch
from torch import nn

from unipercept.nn.losses.mixins import ScaledLossMixin
from unipercept.types import Tensor

from ._stable import EPS_FP32, div, log, sqrt

#######################
# Utilities for depth #
#######################


def depth_to_normals(depth: Tensor, fx: float, fy: float) -> Tensor:
    r"""Compute surface normals from depth map."""
    # Compute gradients in x and y directions
    depth_dx = torch.nn.functional.conv2d(
        depth, torch.tensor([[[[0, 0, 0], [-1, 0, 1], [0, 0, 0]]]]).to(depth.device)
    )
    depth_dy = torch.nn.functional.conv2d(
        depth, torch.tensor([[[[0, -1, 0], [0, 0, 0], [0, 1, 0]]]]).to(depth.device)
    )

    # Compute surface normals
    normals = torch.stack(
        [
            -depth_dx * fx,
            -depth_dy * fy,
            torch.ones_like(depth),
        ],
        dim=-1,
    )

    # Normalize normals
    return torch.nn.functional.normalize(normals, p=2, dim=-1)


def depth_to_disp(
    depth: Tensor,
    depth_min: float,
    depth_max: float,
    fx: float = 1.0,
    eps: float = EPS_FP32,
) -> Tensor:
    r"""Convert depth map to disparity map."""
    return torch.where(depth > eps, fx / depth, torch.zeros_like(depth))


class AvgStat(E.StrEnum):
    r"""Enum for the average statistics available in this module."""

    MEAN = E.auto()
    MEDIAN = E.auto()
    MODE = E.auto()


class VarStat(E.StrEnum):
    r"""Enum for the variance statistics available in this module."""

    VAR = E.auto()
    MAD = E.auto()
    IQR = E.auto()


def _mean_var_with_mask(
    data: Tensor,
    mask: Tensor | None,
    *,
    dim: tuple[int, ...] | int | str,
    keepdim: bool = False,
    correction: int = 1,
) -> tuple[Tensor, Tensor]:
    r"""
    Compute the mean and variance of the data tensor along the specified dimension,
    optionally using a mask tensor.
    """
    if mask is None:
        return data.mean(dim=dim, keepdim=True), data.var(
            dim=dim, keepdim=keepdim, correction=correction
        )
    mask = mask.float()
    mask_sum = torch.sum(mask, dim=dim, keepdim=True)
    mask_mean = torch.sum(data * mask, dim=dim, keepdim=True) / torch.clamp(
        mask_sum, min=1.0
    )
    mask_var = torch.sum(
        mask * (data - mask_mean) ** 2, dim=dim, keepdim=True
    ) / torch.clamp(mask_sum - correction, min=1.0)
    if not keepdim:
        mask_mean = mask_mean.squeeze(dim)
        mask_var = mask_var.squeeze(dim)

    return mask_mean, mask_var


def _mean_with_mask(
    data: Tensor,
    mask: Tensor | None,
    *,
    dim: tuple[int, ...] | int | str,
    keepdim: bool = False,
) -> Tensor:
    r"""
    Compute the mean of the data tensor along the specified dimension, optionally
    using a mask tensor.
    """
    if mask is None:
        return data.mean(dim=dim, keepdim=keepdim)
    mask = mask.float()
    mask_sum = torch.sum(mask, dim=dim, keepdim=True)
    mask_mean = torch.sum(data * mask, dim=dim, keepdim=True) / torch.clamp(
        mask_sum, min=1.0
    )
    if not keepdim:
        mask_mean = mask_mean.squeeze(dim)
    return mask_mean


def _median_with_mask(
    data: Tensor, mask: Tensor | None, *, dim: int | str, keepdim: bool = False
) -> Tensor:
    r"""
    Compute the median of the data tensor along the specified dimension, optionally
    using a mask tensor.
    """
    if mask is None:
        return data.median(dim=dim, keepdim=keepdim).values
    data = torch.where(mask, data, torch.nan)
    return data.nanmedian(dim=dim, keepdim=keepdim).values


def _median_mad_with_mask(
    data: Tensor, mask: Tensor | None, *, dim: int | str, keepdim: bool = False
) -> tuple[Tensor, Tensor]:
    """
    Compute the median and mean absolute deviation of the data tensor along the specified
    dimension, optionally using a mask tensor.
    """
    if mask is None:
        median = data.median(dim=dim, keepdim=keepdim).values
        mad = (data - median).abs().median(dim=dim, keepdim=keepdim).values
        return median, mad
    data = torch.where(mask, data, torch.nan)
    median = data.nanmedian(dim=dim, keepdim=True).values
    mad = (data - median).abs().nanmedian(dim=dim, keepdim=True).values
    if not keepdim:
        median = median.squeeze(dim)
        mad = mad.squeeze(dim)
    return median, mad


####################################
# Scale-Invariant Logarithmic Loss #
####################################


def compute_silog_loss(
    src: Tensor,
    tgt: Tensor,
    mask: Tensor | None,
    dim: tuple[int, ...],
    alpha: float = 0.15,
    eps: float = EPS_FP32,
):
    r"""
    Compute the Scale-Invariant Logarithmic (SILog) loss between the source and target.

    Parameters
    ----------
    src : Tensor[N, ..., *dim]
        The source tensor.
    tgt : Tensor[N, ..., *dim]
        The target tensor.
    mask : Tensor[N, ..., *dim]
        The mask tensor.
    dim : Tuple[int, ...]
        The dimensions to reduce over.
    alpha : float
        The scale factor for the error variance. By default 0.15.
    eps : float
        Small floating point value added for numeric stability.
    """
    err = log(src, eps=eps) - log(tgt, eps=eps)
    err_mean, err_var = _mean_var_with_mask(err, mask, dim=dim, correction=0)
    err_scale = err_mean.square()
    err_scale = alpha * err_scale
    return sqrt(err_var + err_scale, eps=eps)


class SILogLoss(ScaledLossMixin, nn.Module):
    alpha: T.Final[float]
    dim: T.Final[tuple[int, ...]]
    eps: T.Final[float]

    def __init__(
        self,
        alpha: float = 0.15,
        eps: float = EPS_FP32,
        dim: tuple[int, ...] = (-2, -1),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.alpha: float = alpha
        self.dim = dim
        self.eps: float = eps

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(
        self,
        input: Tensor,
        target: Tensor,
        mask: Tensor,
    ) -> Tensor:
        r"""See :func:`compute_silog_loss` for more details."""
        loss = compute_silog_loss(
            input, target, mask, dim=self.dim, alpha=self.alpha, eps=self.eps
        )
        return loss * self.scale


class LogRatioLoss(ScaledLossMixin, nn.Module):
    r"""
    Log Ratio (LogRatio) loss for depth estimation, which is a symmetric approximation
    to the relative error.
    """

    dim: T.Final[tuple[int, ...]]
    eps: T.Final[float]
    smooth: T.Final[bool]

    def __init__(
        self,
        dim: tuple[int, ...] = (-2, -1),
        eps: float = EPS_FP32,
        smooth: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.dim = dim
        self.eps = eps
        self.smooth = smooth

    @T.override
    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        mask: Tensor,
    ) -> Tensor:
        r"""
        Computes the Log Ratio (LogRatio) loss. This shared many of the same properties as
        the Symmetric Relative loss, but is computed using the log of the ratio of the
        source and target values.
        """
        err = torch.log((src + self.eps) / (tgt + self.eps))
        if self.smooth:
            err = log_cosh(err)
        return _mean_with_mask(data=err, mask=mask, dim=self.dim) * self.scale


class AbsRelLoss(ScaledLossMixin, nn.Module):
    r"""Relative loss for depth estimation."""

    dim: T.Final[tuple[int, ...]]
    eps: T.Final[float]
    symmetric: T.Final[bool]
    tangent: T.Final[bool]

    def __init__(
        self,
        dim: tuple[int, ...] = (-2, -1),
        eps: float = EPS_FP32,
        symmetric: bool = False,
        tangent: bool = False,
        **kwargs,
    ) -> None:
        r"""
        Parameters
        ----------
        dim : Tuple[int, ...]
            The dimensions to reduce over.
        eps : float
            Small floating point value added for numeric stability, used in the
            denominator.
        symmetric : bool
            If True, the loss is computed symmetrically. This implies a maximum range
            of [0, 2].
        tangent : bool
            If True, the loss is computed using the arctangent of the relative error.
            This is useful for better handling of large errors.
        """
        super().__init__(**kwargs)
        self.dim = dim
        self.eps = eps
        self.symmetric = symmetric
        self.tangent = tangent

    def compute(self, src: Tensor, tgt: Tensor) -> Tensor:
        return nn.functional.l1_loss(src, tgt, reduction="none")

    @T.override
    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        mask: Tensor,
    ) -> Tensor:
        """See :func:`compute_absrel_loss` for more details."""
        err = self.compute(src, tgt)
        if self.symmetric:
            err = 2 * err
            tgt = src + tgt
        err = div(err, tgt, eps=self.eps)
        if self.tangent:
            err = torch.atan(err)
        return _mean_with_mask(data=err, mask=mask, dim=self.dim) * self.scale


class SqRelLoss(AbsRelLoss):
    @T.override
    def compute(self, src: Tensor, tgt: Tensor) -> Tensor:
        return nn.functional.mse_loss(src, tgt, reduction="none")


class SmoothRelLoss(AbsRelLoss):
    beta: T.Final[float]

    @T.override
    def compute(self, src: Tensor, tgt: Tensor) -> Tensor:
        return nn.functional.smooth_l1_loss(src, tgt, beta=self.beta, reduction="none")

    def __init__(
        self,
        *args,
        beta: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.beta = beta


class HuberRelLoss(AbsRelLoss):
    delta: T.Final[float]

    @T.override
    def compute(self, src: Tensor, tgt: Tensor) -> Tensor:
        return nn.functional.huber_loss(src, tgt, delta=self.delta, reduction="none")

    def __init__(
        self,
        *args,
        delta: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.delta = delta


def log_cosh(x: Tensor) -> Tensor:
    r"""Numerically stable Log-Cosh function."""
    return (
        x
        + torch.nn.functional.softplus(-2.0 * x, beta=1.0, threshold=20.0)
        - math.log(2.0)
    )


class LogCoshRelLoss(AbsRelLoss):
    @T.override
    def compute(self, src: Tensor, tgt: Tensor) -> Tensor:
        return log_cosh(nn.functional.l1_loss(src, tgt, reduction="none"))


_DEFAULT_TWEEDIE_P: T.Final[float] = 1.5


def tweedie_loss(
    src: Tensor,
    tgt: Tensor,
    mask: Tensor | None,
    dim: tuple[int, ...],
    p: float = _DEFAULT_TWEEDIE_P,
) -> Tensor:
    """Computes the Tweedie loss."""
    a = tgt * torch.exp(src * (1 - p)) / (1 - p)
    b = torch.exp(src * (2 - p)) / (2 - p)
    err = (b - a).abs()

    err = _mean_with_mask(data=err, mask=mask, dim=dim)
    if err.ndim > 1:
        err = err.sum(dim=1)
    return err


class TweedieLoss(nn.Module):
    dim: T.Final[tuple[int, ...]]
    p: T.Final[float]

    def __init__(
        self,
        *args,
        dim: tuple[int, ...] = (-2, -1),
        p: float = _DEFAULT_TWEEDIE_P,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.p = p
        self.dim = dim

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        mask: Tensor | None,
    ) -> Tensor:
        """See :func:`tweedie_loss` for more details."""
        return tweedie_loss(src, tgt, mask, dim=self.dim, p=self.p) * self.scale


def l1_loss(
    src: Tensor,
    tgt: Tensor,
    mask: Tensor,
    dim: tuple[int, ...],
):
    r"""
    Compute the Mean Squared Error (L1) loss between the source and target
    tensors.

    Parameters
    ----------
    src : Tensor[N, ..., *dim]
        The source tensor.
    tgt : Tensor[N, ..., *dim] or Tensor[N, ..., *dim, P]
        The target tensor.
    mask : Tensor[N, ..., *dim]
        The mask tensor.
    dim : Tuple[int, ...]
        The dimensions to reduce over.
    eps : float
        Small floating point value added for numeric stability.
    """
    err = torch.nn.functional.l1_loss(src, tgt, reduction="none")
    return _mean_with_mask(err, mask=mask, dim=dim)


class L1Loss(ScaledLossMixin, nn.Module):
    dim: T.Final[tuple[int, ...]]

    def __init__(
        self,
        dim: tuple[int, ...] = (-2, -1),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.dim = dim

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        mask: Tensor,
    ) -> Tensor:
        """See :func:`compute_l1_loss` for more details."""
        return l1_loss(src, tgt, mask, dim=self.dim) * self.scale


def mse_loss(
    src: Tensor,
    tgt: Tensor,
    mask: Tensor,
    dim: tuple[int, ...],
):
    r"""
    Compute the Mean Squared Error (MSE) loss between the source and target
    tensors.

    Parameters
    ----------
    src : Tensor[N, ..., *dim]
        The source tensor.
    tgt : Tensor[N, ..., *dim] or Tensor[N, ..., *dim, P]
        The target tensor.
    mask : Tensor[N, ..., *dim]
        The mask tensor.
    dim : Tuple[int, ...]
        The dimensions to reduce over.
    """
    err = torch.nn.functional.mse_loss(src, tgt, reduction="none")
    return _mean_with_mask(err, mask=mask, dim=dim)


class MSELoss(L1Loss):
    """See :func:`compute_mse_loss` for more details."""

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        mask: Tensor,
    ) -> Tensor:
        return mse_loss(src, tgt, mask, dim=self.dim) * self.scale


def rmse_loss(
    src: Tensor,
    tgt: Tensor,
    mask: Tensor,
    dim: tuple[int, ...],
    eps: float = EPS_FP32,
):
    r"""
    Compute the Root Mean Squared Error (RMSE) loss between the source and target
    tensors.

    Parameters
    ----------
    src : Tensor[N, ..., *dim]
        The source tensor.
    tgt : Tensor[N, ..., *dim] or Tensor[N, ..., *dim, P]
        The target tensor.
    mask : Tensor[N, ..., *dim]
        The mask tensor.
    dim : Tuple[int, ...]
        The dimensions to reduce over.
    eps : float
        Small floating point value added for numeric stability.
    """
    err = torch.nn.functional.mse_loss(src, tgt, reduction="none")
    err = sqrt(err, eps=eps)
    return _mean_with_mask(err, mask=mask, dim=dim)


class RMSELoss(L1Loss):
    """See :func:`compute_mse_loss` for more details."""

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        mask: Tensor,
    ) -> Tensor:
        return rmse_loss(src, tgt, mask, dim=self.dim) * self.scale


class DCELoss(ScaledLossMixin, nn.Module):
    """Discretized Cross-Entropy (DCE) loss for Depth Estimation."""

    def __init__(
        self,
        depth_min: float = 1.0,
        depth_max: float = 87.0,
        out_channel=200,
        eps=EPS_FP32,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.depth_min = depth_min
        self.depth_max = depth_max
        self.bins_num = out_channel
        self.alpha = 2
        self.noise_sample_ratio = 0.9
        self.eps = eps
        self.depth_bins_interval = (
            torch.log10(torch.tensor(self.depth_max)) - math.log10(self.depth_min)
        ) / self.bins_num
        bins_edges_in_log = math.log10(
            self.depth_min
        ) + self.depth_bins_interval * torch.tensor(
            [*list(range(self.bins_num)), self.bins_num]
        )
        bins_weight = torch.tensor(
            [
                [math.exp(-self.alpha * (i - j) ** 2) for i in range(self.bins_num)]
                for j in range(self.bins_num)
            ]
        )
        self.register_buffer("bins_weight", bins_weight.float(), persistent=False)
        self.register_buffer(
            "bins_edges_in_log", bins_edges_in_log.float(), persistent=False
        )

    def depth_to_bins_in_log(self, depth, mask):
        """
        Discretize depth into depth bins. Predefined bins edges are in log space.
        Mark invalid padding area as bins_num + 1.

        Parameters
        ----------
        depth: Tensor[N, 1, H, W]
            The depth tensor.
        mask: Tensor[N, 1, H, W]
            The mask tensor.

        Returns
        -------
        depth_bins_log: Tensor[N, C, H, W]
            The depth bins.
        """
        invalid_mask = ~mask
        # depth[depth < self.depth_min] = self.depth_min
        # depth[depth > self.depth_max] = self.depth_max
        mask_lower = depth <= self.depth_min
        mask_higher = depth >= self.depth_max
        depth_bins_log = (
            (torch.log10(torch.abs(depth)) - math.log10(self.depth_min))
            / self.depth_bins_interval
        ).to(torch.int)

        depth_bins_log[mask_lower] = 0
        depth_bins_log[mask_higher] = self.bins_num - 1
        depth_bins_log[depth_bins_log == self.bins_num] = self.bins_num - 1

        depth_bins_log[invalid_mask] = self.bins_num + 1
        return depth_bins_log

    def depth_to_bins(self, depth, mask, depth_edges, size_bounds=(300, 300)):
        """
        Discretize depth into depth bins. Predefined bins edges are provided.
        Mark invalid padding area as bins_num + 1.

        Parameters
        ----------
        depth: Tensor[N, 1, H, W]
            The depth tensor.
        mask: Tensor[N, 1, H, W]
            The mask tensor.
        depth_edges: Tensor[C]
            The predefined depth bins edges.
        size_bounds: Tuple[int, int]
            The size limit of the block to process.

        Returns
        -------
        Tensor[N, C, H, W]
            The depth discretized into bins.
        """

        def _depth_to_bins_block_(depth, mask, depth_edges):
            bins_id = torch.sum(
                depth_edges[:, None, None, None, :]
                < torch.abs(depth)[:, :, :, :, None],
                dim=-1,
            )
            bins_id = bins_id - 1
            invalid_mask = ~mask
            mask_lower = depth <= self.depth_min
            mask_higher = depth >= self.depth_max

            bins_id[mask_lower] = 0
            bins_id[mask_higher] = self.bins_num - 1
            bins_id[bins_id == self.bins_num] = self.bins_num - 1

            bins_id[invalid_mask] = self.bins_num + 1
            return bins_id

        _, _, H, W = depth.shape
        bins = mask.clone().long()
        h_blocks = math.ceil(H / size_bounds[0])
        w_blocks = math.ceil(W / size_bounds[1])
        for i in range(h_blocks):
            for j in range(w_blocks):
                h_start = i * size_bounds[0]
                h_end_proposal = (i + 1) * size_bounds[0]
                h_end = min(H, h_end_proposal)
                w_start = j * size_bounds[1]
                w_end_proposal = (j + 1) * size_bounds[1]
                w_end = min(W, w_end_proposal)
                bins_ij = _depth_to_bins_block_(
                    depth[:, :, h_start:h_end, w_start:w_end],
                    mask[:, :, h_start:h_end, w_start:w_end],
                    depth_edges,
                )
                bins[:, :, h_start:h_end, w_start:w_end] = bins_ij
        return bins

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(self, src_logits: Tensor, tgt_depths: Tensor, mask: Tensor, **kwargs):
        B, _, _H, _W = tgt_depths.shape
        if "bins_edges" not in kwargs or kwargs["bins_edges"] is None:
            # predefined depth bins in log space
            tgt_bins = self.depth_to_bins_in_log(tgt_depths, mask).float()
        else:
            bins_edges = kwargs["bins_edges"]
            tgt_bins = self.depth_to_bins(tgt_depths, mask, bins_edges).float()

        classes_range = torch.arange(
            self.bins_num, device=tgt_bins.device, dtype=tgt_bins.dtype
        )
        log_pred = torch.nn.functional.log_softmax(src_logits, 1)
        log_pred = log_pred.reshape(B, log_pred.size(1), -1).permute((0, 2, 1))
        gt_reshape = tgt_bins.reshape((B, -1))[:, :, None]
        one_hot = (gt_reshape == classes_range).to(
            dtype=torch.float, device=src_logits.device
        )
        weight = torch.matmul(one_hot, self.bins_weight)
        weight_log_pred = weight * log_pred
        loss_pixeles = -torch.sum(weight_log_pred, dim=2)

        valid_pixels = torch.sum(mask).to(dtype=torch.float, device=src_logits.device)
        loss = torch.sum(loss_pixeles) / valid_pixels.clamp_min(1.0)
        return loss * self.scale


class HDSNRandomLoss(nn.Module):
    """
    Hieratical depth spatial normalization loss.
    Replace the original grid masks with the random created masks.

    .. math::
        L = MAE((d-median(d)/s - (d'-median(d'))/s'), s = mean(d- median(d))

    See Also
    --------
    - `Paper <https://arxiv.org/pdf/2404.15506>`_
    """

    def __init__(
        self, scale=1.0, random_num=32, sky_id=142, batch_limit=8, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.scale = scale
        self.random_num = random_num
        self.sky_id = sky_id
        self.batch_limit = batch_limit
        self.eps = 1e-6
        self.disable_dataset = []

    def get_random_masks_for_batch(
        self, image_size: list, device: torch.device
    ) -> Tensor:
        height, width = image_size
        crop_h_min = int(0.125 * height)
        crop_h_max = int(0.5 * height)
        crop_w_min = int(0.125 * width)
        crop_w_max = int(0.5 * width)
        h_max = height - crop_h_min
        w_max = width - crop_w_min
        crop_height = (
            torch.randperm(crop_h_max - crop_h_min, device=device)[: self.random_num]
            + crop_h_min
        )
        crop_width = (
            torch.randperm(crop_w_max - crop_w_min, device=device)[: self.random_num]
            + crop_w_min
        )
        crop_y = torch.randperm(h_max, device=device)[: self.random_num]
        crop_x = torch.randperm(w_max, device=device)[: self.random_num]
        crop_y_end = crop_height + crop_y
        crop_y_end[crop_y_end >= height] = height
        crop_x_end = crop_width + crop_x
        crop_x_end[crop_x_end >= width] = width

        mask_new = torch.zeros(
            (self.random_num, height, width), dtype=torch.bool, device=device
        )  # .cuda() #[N, H, W]
        for i in range(self.random_num):
            mask_new[i, crop_y[i] : crop_y_end[i], crop_x[i] : crop_x_end[i]] = True

        return mask_new
        # return crop_y, crop_y_end, crop_x, crop_x_end

    def reorder_sem_masks(self, sem_label):
        # reorder the semantic mask of a batch
        assert sem_label.ndim == 3
        semantic_ids = torch.unique(
            sem_label[(sem_label > 0) & (sem_label != self.sky_id)]
        )
        sem_masks = [sem_label == id for id in semantic_ids]
        if len(sem_masks) == 0:
            # no valid semantic labels
            return sem_label > 0

        sem_masks = torch.cat(sem_masks, dim=0)
        mask_batch = torch.sum(sem_masks.reshape(sem_masks.shape[0], -1), dim=1) > 500
        sem_masks = sem_masks[mask_batch]
        if sem_masks.shape[0] > self.random_num:
            balance_samples = torch.randperm(
                sem_masks.shape[0], device=sem_label.device
            )[: self.random_num]
            sem_masks = sem_masks[balance_samples, ...]

        if sem_masks.shape[0] == 0:
            # no valid semantic labels
            return sem_label > 0

        if sem_masks.ndim == 2:
            sem_masks = sem_masks[None, :, :]
        return sem_masks

    def ssi_mae(self, prediction, target, mask_valid):
        B, C, _H, _W = target.shape
        prediction_nan = prediction.clone().detach()
        target_nan = target.clone()
        prediction_nan[~mask_valid] = float("nan")
        target_nan[~mask_valid] = float("nan")

        valid_pixs = mask_valid.reshape((B, C, -1)).sum(dim=2, keepdims=True) + 1e-10
        valid_pixs = valid_pixs[:, :, :, None]

        gt_median = (
            target_nan.reshape((B, C, -1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1)
        )  # [b,c,h,w]
        gt_median[torch.isnan(gt_median)] = 0
        gt_diff = (torch.abs(target - gt_median)).reshape((B, C, -1))
        gt_s = gt_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        gt_trans = (target - gt_median) / (gt_s + self.eps)

        pred_median = (
            prediction_nan.reshape((B, C, -1))
            .nanmedian(2, keepdims=True)[0]
            .unsqueeze(-1)
        )  # [b,c,h,w]
        pred_median[pred_median.isnan()] = 0
        pred_diff = (torch.abs(prediction - pred_median)).reshape((B, C, -1))
        pred_s = pred_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        pred_trans = (prediction - pred_median) / (pred_s + self.eps)

        return torch.sum(torch.abs(gt_trans - pred_trans) * mask_valid)

    def conditional_ssi_mae(self, prediction, target, mask_valid):
        B, C, _H, _W = target.shape
        conditional_rank_ids = torch.randperm(B, device=target.device)

        prediction_nan = prediction.clone()
        target_nan = target.clone()
        prediction_nan[~mask_valid] = float("nan")
        target_nan[~mask_valid] = float("nan")

        valid_pixs = mask_valid.reshape((B, C, -1)).sum(dim=2, keepdims=True) + self.eps
        valid_pixs = valid_pixs[:, :, :, None].contiguous()

        gt_median = (
            target_nan.reshape((B, C, -1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1)
        )  # [b,c,h,w]
        gt_median[torch.isnan(gt_median)] = 0
        gt_diff = (torch.abs(target - gt_median) * mask_valid).reshape((B, C, -1))
        gt_s = gt_diff.sum(dim=2)[:, :, None, None].contiguous() / valid_pixs

        # in case some batches have no valid pixels
        gt_s_small_mask = gt_s < (torch.mean(gt_s) * 0.1)
        gt_s[gt_s_small_mask] = torch.mean(gt_s)
        gt_trans = (target - gt_median[conditional_rank_ids]) / (
            gt_s[conditional_rank_ids] + self.eps
        )

        pred_median = (
            prediction_nan.reshape((B, C, -1))
            .nanmedian(2, keepdims=True)[0]
            .unsqueeze(-1)
        )  # [b,c,h,w]
        pred_median[torch.isnan(pred_median)] = 0
        pred_diff = (torch.abs(prediction - pred_median) * mask_valid).reshape(
            (B, C, -1)
        )
        pred_s = pred_diff.sum(dim=2)[:, :, None, None].contiguous() / valid_pixs
        pred_s[gt_s_small_mask] = torch.mean(pred_s)
        pred_trans = (prediction - pred_median[conditional_rank_ids]) / (
            pred_s[conditional_rank_ids] + self.eps
        )

        return torch.sum(torch.abs(gt_trans - pred_trans) * mask_valid)
        # print(torch.abs(gt_trans - pred_trans)[mask_valid])

    def forward(self, src, tgt, mask=None, sem_mask=None, **kwargs):
        """Calculate loss."""
        B, _C, H, W = tgt.shape

        loss = 0.0
        valid_pix = 0.0

        device = tgt.device

        batches_dataset = kwargs["dataset"]
        self.batch_valid = torch.tensor(
            [
                1 if batch_dataset not in self.disable_dataset else 0
                for batch_dataset in batches_dataset
            ],
            device=device,
        )[:, None, None, None]

        batch_limit = self.batch_limit

        random_sample_masks = self.get_random_masks_for_batch(
            (H, W), device=device
        )  # [N, H, W]

        # Batch valid mask
        bv = self.batch_valid.squeeze()  # [B]
        if bv.ndim == 0:
            bv = bv.unsqueeze(0)

        for i in range(B):
            # Skip if invalid
            if bv[i] < 0.5:
                loss += 0 * src[i].sum()
                valid_pix += 0 * mask[i].sum()
                continue

            mask_i = mask[i]  # [1, H, W]
            pred_i = src[i].unsqueeze(0)  # [1, C, H, W]
            target_i = tgt[i].unsqueeze(0)  # [1, C, H, W]

            # get semantic masks
            sem_label_i = sem_mask[i] if sem_mask is not None else None
            if sem_label_i is not None:
                sem_masks = self.reorder_sem_masks(sem_label_i)  # [N, H, W]
                # [N_rnd + N_sem, H, W]
                combined_masks = torch.cat([random_sample_masks, sem_masks], dim=0)
            else:
                combined_masks = random_sample_masks

            n_masks = combined_masks.shape[0]

            # Use chunks to limit memory usage
            for j in range(0, n_masks, batch_limit):
                chunk_masks = combined_masks[j : j + batch_limit]
                n_chunk = chunk_masks.shape[0]

                # Combine with global mask [N_chunk, 1, H, W]
                mask_sample = (mask_i & chunk_masks).unsqueeze(1)

                # Repeat prediction and target for the chunk
                p_chunk = pred_i.expand(n_chunk, -1, -1, -1)
                t_chunk = target_i.expand(n_chunk, -1, -1, -1)

                loss += self.ssi_mae(
                    prediction=p_chunk,
                    target=t_chunk,
                    mask_valid=mask_sample,
                )
                valid_pix += mask_sample.sum()

        # image-wide SSI loss
        img_mask = mask * self.batch_valid.bool()
        loss += self.ssi_mae(prediction=src, target=tgt, mask_valid=img_mask)
        valid_pix += img_mask.sum()
        loss = loss / (valid_pix + self.eps)
        if torch.isnan(loss).item() | torch.isinf(loss).item():
            loss = 0 * torch.sum(src)
            print(f"HDSNL NAN error, {loss}, valid pix: {valid_pix}")
        return loss * self.scale


class HDNRandomLoss(nn.Module):
    """
    Hieratical depth normalization loss.
    Replace the original hieratical depth ranges with randomly sampled ranges.

    .. math::

        L = MAE((d-median(d)/s - (d'-median(d'))/s'), s = mean(d- median(d))

    See Also
    --------
    - `Paper <https://arxiv.org/pdf/2404.15506>`_
    """

    def __init__(self, scale=1, random_num=32, eps=EPS_FP32, **kwargs) -> None:
        super().__init__()
        self.scale = scale
        self.random_num = random_num
        self.eps = eps
        self.disable_dataset = []

    def get_random_masks_for_batch(
        self, depth_gt: Tensor, mask_valid: Tensor
    ) -> Tensor:
        valid_values = depth_gt[mask_valid]
        max_d = valid_values.max().item() if valid_values.numel() > 0 else 0.0
        min_d = valid_values.min().item() if valid_values.numel() > 0 else 0.0

        sample_min_d = [
            random.uniform(0, 0.75) * (max_d - min_d) + min_d
            for _ in range(self.random_num)
        ]
        sample_max_d = [
            random.uniform(sample_min_d[i] + 0.1, 1 - self.eps) * (max_d - min_d)
            + min_d
            for i in range(self.random_num)
        ]

        mask_new = [
            (depth_gt >= sample_min_d[i])
            & (depth_gt < sample_max_d[i] + 1e-30)
            & mask_valid
            for i in range(self.random_num)
        ]
        return torch.stack(mask_new, dim=0).to(depth_gt.device)  # [N, 1, H, W]

    def ssi_mae(self, prediction, target, mask_valid):
        B, C, _H, _W = target.shape
        prediction_nan = prediction.clone().detach()
        target_nan = target.clone()
        prediction_nan[~mask_valid] = float("nan")
        target_nan[~mask_valid] = float("nan")

        valid_pixs = mask_valid.reshape((B, C, -1)).sum(dim=2, keepdims=True) + self.eps
        valid_pixs = valid_pixs[:, :, :, None]

        gt_median = (
            target_nan.reshape((B, C, -1)).nanmedian(2, keepdims=True)[0].unsqueeze(-1)
        )  # [b,c,h,w]
        gt_median[torch.isnan(gt_median)] = 0
        gt_diff = (torch.abs(target - gt_median) * mask_valid).reshape((B, C, -1))
        gt_s = gt_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        gt_trans = (target - gt_median) / (gt_s + self.eps)

        pred_median = (
            prediction_nan.reshape((B, C, -1))
            .nanmedian(2, keepdims=True)[0]
            .unsqueeze(-1)
        )  # [b,c,h,w]
        pred_median[torch.isnan(pred_median)] = 0
        pred_diff = (torch.abs(prediction - pred_median) * mask_valid).reshape(
            (B, C, -1)
        )
        pred_s = pred_diff.sum(dim=2)[:, :, None, None] / valid_pixs
        pred_trans = (prediction - pred_median) / (pred_s + self.eps)

        return torch.sum(torch.abs(gt_trans - pred_trans) * mask_valid)

    @T.override
    def forward(self, prediction, target, mask=None, **kwargs):
        """Calculate loss."""
        B, _C, _H, _W = target.shape

        loss = 0.0
        valid_pix = 0.0

        device = target.device

        batches_dataset = kwargs["dataset"]
        self.batch_valid = torch.tensor(
            [
                1 if batch_dataset not in self.disable_dataset else 0
                for batch_dataset in batches_dataset
            ],
            device=device,
        )[:, None, None, None]

        batch_limit = 4
        loops = int(torch.ceil(self.random_num / batch_limit))
        for i in range(B):
            mask_i = mask[i, ...]  # [1, H, W]

            if self.batch_valid[i, ...] < 0.5:
                loss += 0 * torch.sum(prediction[i, ...])
                valid_pix += 0 * torch.sum(mask_i)
                continue

            pred_i = prediction[i, ...].unsqueeze(0).repeat(batch_limit, 1, 1, 1)
            target_i = target[i, ...].unsqueeze(0).repeat(batch_limit, 1, 1, 1)
            mask_random_drange = self.get_random_masks_for_batch(
                target[i, ...], mask_i
            )  # [N, 1, H, W]
            for j in range(loops):
                mask_random_loopi = mask_random_drange[
                    j * batch_limit : (j + 1) * batch_limit, ...
                ]
                loss += self.ssi_mae(
                    prediction=pred_i[: mask_random_loopi.shape[0], ...],
                    target=target_i[: mask_random_loopi.shape[0], ...],
                    mask_valid=mask_random_loopi,
                )
                valid_pix += torch.sum(mask_random_loopi)

        loss = loss / valid_pix.clamp_min(1.0)
        return loss * self.scale


class VNLoss(ScaledLossMixin, nn.Module):
    """
    Virtual Normal Loss.

    See Also
    --------
    - `Paper <https://arxiv.org/pdf/2103.04216>`_
    """

    def __init__(
        self,
        delta_cos=0.867,
        delta_diff_x=0.01,
        delta_diff_y=0.01,
        delta_diff_z=0.01,
        delta_z=1e-5,
        sample_ratio=0.2,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.delta_cos = delta_cos
        self.delta_diff_x = delta_diff_x
        self.delta_diff_y = delta_diff_y
        self.delta_diff_z = delta_diff_z
        self.delta_z = delta_z
        self.sample_ratio = sample_ratio
        self.eps = 1e-6

    def init_image_coor(self, intrinsic, height, width) -> None:
        u0 = intrinsic[:, 0, 2][:, None, None, None]
        v0 = intrinsic[:, 1, 2][:, None, None, None]
        device = intrinsic.device
        y, x = torch.meshgrid(
            [
                torch.arange(0, height, dtype=torch.float32, device=device),
                torch.arange(0, width, dtype=torch.float32, device=device),
            ],
            indexing="ij",
        )
        u_m_u0 = x[None, None, :, :] - u0
        v_m_v0 = y[None, None, :, :] - v0
        # return u_m_u0, v_m_v0
        self.register_buffer("v_m_v0", v_m_v0, persistent=False)
        self.register_buffer("u_m_u0", u_m_u0, persistent=False)

    def transfer_xyz(self, depth, focal_length, u_m_u0, v_m_v0):
        x = u_m_u0 * depth / focal_length
        y = v_m_v0 * depth / focal_length
        z = depth
        return torch.cat([x, y, z], 1).permute(0, 2, 3, 1).contiguous()  # [b, h, w, c]

    def select_index(self, B, H, W, mask):
        device = mask.device
        intend_sample_num = int(H * W * self.sample_ratio)

        # Pre-calculate pixel indices
        pix_idx_mat = (
            torch.arange(H * W, device=device).reshape((1, H, W)).expand(B, H, W)
        )

        # Prepare outputs
        p1 = torch.zeros((B, intend_sample_num), dtype=torch.long, device=device)
        p2 = torch.zeros((B, intend_sample_num), dtype=torch.long, device=device)
        p3 = torch.zeros((B, intend_sample_num), dtype=torch.long, device=device)

        for i in range(B):
            # Mask-based selection per batch element
            mask_i = mask[i, 0] > self.eps
            inputs_index = pix_idx_mat[i][mask_i]
            num_effect_pixels = inputs_index.numel()

            if num_effect_pixels > 0:
                sample_num = min(intend_sample_num, num_effect_pixels)

                # Sample 3 sets of indices
                idx1 = torch.randperm(num_effect_pixels, device=device)[:sample_num]
                idx2 = torch.randperm(num_effect_pixels, device=device)[:sample_num]
                idx3 = torch.randperm(num_effect_pixels, device=device)[:sample_num]

                p1[i, :sample_num] = inputs_index[idx1]
                p2[i, :sample_num] = inputs_index[idx2]
                p3[i, :sample_num] = inputs_index[idx3]

        return {
            "p1_x": p1 % W,
            "p1_y": torch.div(p1, W, rounding_mode="trunc"),
            "p2_x": p2 % W,
            "p2_y": torch.div(p2, W, rounding_mode="trunc"),
            "p3_x": p3 % W,
            "p3_y": torch.div(p3, W, rounding_mode="trunc"),
        }

    def form_pw_groups(self, p123, pw):
        """
        Form 3D points groups, with 3 points in each grouup.
        :param p123: points index
        :param pw: 3D points
        :return:
        """
        B, _H, _W, _C = pw.shape
        p1_x, p1_y = p123["p1_x"], p123["p1_y"]
        p2_x, p2_y = p123["p2_x"], p123["p2_y"]
        p3_x, p3_y = p123["p3_x"], p123["p3_y"]

        # Efficiently gather 3D points using advanced indexing
        batch_indices = torch.arange(B, device=pw.device)[:, None]
        pw1 = pw[batch_indices, p1_y, p1_x, :]
        pw2 = pw[batch_indices, p2_y, p2_x, :]
        pw3 = pw[batch_indices, p3_y, p3_x, :]

        # [B, N, 3(x,y,z), 3(p1,p2,p3)]
        return torch.stack([pw1, pw2, pw3], dim=3)

    def filter_mask(
        self,
        p123,
        gt_xyz,
        delta_cos=0.867,
        delta_diff_x=0.005,
        delta_diff_y=0.005,
        delta_diff_z=0.005,
    ):
        pw = self.form_pw_groups(p123, gt_xyz)
        pw12 = pw[:, :, :, 1] - pw[:, :, :, 0]
        pw13 = pw[:, :, :, 2] - pw[:, :, :, 0]
        pw23 = pw[:, :, :, 2] - pw[:, :, :, 1]
        ###ignore linear
        pw_diff = torch.cat(
            [
                pw12[:, :, :, None],
                pw13[:, :, :, None],
                pw23[:, :, :, None],
            ],
            3,
        )  # [b, n, 3, 3]
        m_batchsize, groups, _coords, index = pw_diff.shape
        proj_query = (
            pw_diff.view(m_batchsize * groups, -1, index).permute(0, 2, 1).contiguous()
        )  # (B* X CX(3)) [bn, 3(p123), 3(xyz)]
        proj_key = pw_diff.contiguous().view(
            m_batchsize * groups, -1, index
        )  # B X  (3)*C [bn, 3(xyz), 3(p123)]
        q_norm = proj_query.norm(2, dim=2)
        nm = torch.bmm(
            q_norm.contiguous().view(m_batchsize * groups, index, 1),
            q_norm.view(m_batchsize * groups, 1, index),
        )  # []
        energy = torch.bmm(
            proj_query, proj_key
        )  # transpose check [bn, 3(p123), 3(p123)]
        norm_energy = energy / (nm + self.eps)
        norm_energy = norm_energy.contiguous().view(m_batchsize * groups, -1)
        mask_cos = (
            torch.sum((norm_energy > delta_cos) + (norm_energy < -delta_cos), 1) > 3
        )  # igonre
        mask_cos = mask_cos.contiguous().view(m_batchsize, groups)
        ##ignore padding and invilid depth
        mask_pad = torch.sum(pw[:, :, 2, :] > self.delta_z, 2) == 3

        ###ignore near
        mask_x = torch.sum(torch.abs(pw_diff[:, :, 0, :]) < delta_diff_x, 2) > 0
        mask_y = torch.sum(torch.abs(pw_diff[:, :, 1, :]) < delta_diff_y, 2) > 0
        mask_z = torch.sum(torch.abs(pw_diff[:, :, 2, :]) < delta_diff_z, 2) > 0

        mask_ignore = (mask_x & mask_y & mask_z) | mask_cos
        mask_near = ~mask_ignore
        mask = mask_pad & mask_near

        return mask, pw

    def select_points_groups(self, gt_depth, pred_depth, intrinsic, mask):
        B, _C, H, W = gt_depth.shape
        focal_length = intrinsic[:, 0, 0][:, None, None, None]
        u_m_u0, v_m_v0 = (
            self.u_m_u0,
            self.v_m_v0,
        )  # self.init_image_coor(intrinsic, height=H, width=W)

        pw_gt = self.transfer_xyz(gt_depth, focal_length, u_m_u0, v_m_v0)
        pw_pred = self.transfer_xyz(pred_depth, focal_length, u_m_u0, v_m_v0)

        p123 = self.select_index(B, H, W, mask)
        # mask:[b, n], pw_groups_gt: [b, n, 3(x,y,z), 3(p1,p2,p3)]
        mask, pw_groups_gt = self.filter_mask(
            p123,
            pw_gt,
            delta_cos=0.867,
            delta_diff_x=0.005,
            delta_diff_y=0.005,
            delta_diff_z=0.005,
        )

        # [b, n, 3, 3]
        pw_groups_pred = self.form_pw_groups(p123, pw_pred)
        pw_groups_pred[pw_groups_pred[:, :, 2, :] == 0] = 0.0001
        mask_broadcast = (
            mask.repeat(1, 9).reshape(B, 3, 3, -1).permute(0, 3, 1, 2).contiguous()
        )
        pw_groups_pred_not_ignore = pw_groups_pred[mask_broadcast].reshape(1, -1, 3, 3)
        pw_groups_gt_not_ignore = pw_groups_gt[mask_broadcast].reshape(1, -1, 3, 3)

        return pw_groups_gt_not_ignore, pw_groups_pred_not_ignore

    @T.override
    def forward(self, prediction, target, mask, intrinsic, select: bool = True):
        # configs for the cameras
        # focal_length = intrinsic[:, 0, 0][:, None, None, None]
        # u0 = intrinsic[:, 0, 2][:, None, None, None]
        # v0 = intrinsic[:, 1, 2][:, None, None, None]
        B, _, H, W = target.shape
        if (
            "u_m_u0" not in self._buffers
            or "v_m_v0" not in self._buffers
            or self.u_m_u0.shape != torch.Size([B, 1, H, W])
            or self.v_m_v0.shape != torch.Size([B, 1, H, W])
        ):
            self.init_image_coor(intrinsic, H, W)

        gt_points, pred_points = self.select_points_groups(
            target, prediction, intrinsic, mask
        )

        gt_p12 = gt_points[:, :, :, 1] - gt_points[:, :, :, 0]
        gt_p13 = gt_points[:, :, :, 2] - gt_points[:, :, :, 0]
        pred_p12 = pred_points[:, :, :, 1] - pred_points[:, :, :, 0]
        pred_p13 = pred_points[:, :, :, 2] - pred_points[:, :, :, 0]

        gt_normal = torch.cross(gt_p12, gt_p13, dim=2)
        pred_normal = torch.cross(pred_p12, pred_p13, dim=2)
        pred_norm = torch.norm(pred_normal, 2, dim=2, keepdim=True)
        gt_norm = torch.norm(gt_normal, 2, dim=2, keepdim=True)
        pred_mask = pred_norm == 0.0
        gt_mask = gt_norm == 0.0
        pred_mask = pred_mask.to(torch.float32)
        gt_mask = gt_mask.to(torch.float32)
        pred_mask *= self.eps
        gt_mask *= self.eps
        gt_norm = gt_norm + gt_mask
        pred_norm = pred_norm + pred_mask
        gt_normal = gt_normal / gt_norm
        pred_normal = pred_normal / pred_norm
        loss = torch.abs(gt_normal - pred_normal)
        loss = torch.sum(torch.sum(loss, dim=2), dim=0)
        if select:
            loss, _indices = torch.sort(loss, dim=0, descending=False)
            loss = loss[int(loss.size(0) * 0.25) :]

        return loss.mean() * self.scale


def mask_discontinuity_loss(src: Tensor, tgt: Tensor) -> Tensor:
    """
    Computes the mask-guided edge discontinuity loss for a given disparity map and
    a binary mask.
    """
    # Compute the Iverson bracket for adjacent pixels
    tgt_dx = torch.diff(tgt, dim=-1) != 0
    tgt_dy = torch.diff(tgt, dim=-2) != 0

    # Compute the partial disparity derivative
    src_dx = torch.diff(src, dim=-1)
    src_dy = torch.diff(src, dim=-2)

    loss_x = torch.mean(torch.mul(tgt_dx, torch.exp(-torch.abs(src_dx))))
    loss_y = torch.mean(torch.mul(tgt_dy, torch.exp(-torch.abs(src_dy))))
    return loss_x + loss_y


class PEDLoss(nn.Module):
    """Mask-Guided Edge Discontinuity Loss (MGED) loss."""

    def __init__(self) -> None:
        super().__init__()

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(self, src: Tensor, tgt: Tensor) -> Tensor:
        src = src.float()
        tgt = tgt.float()
        return mask_discontinuity_loss(src, tgt)


def compute_image_guided_smoothness_loss(src: Tensor, tgt: Tensor) -> Tensor:
    """
    Compute the smoothness loss for a given disparity map.

    Parameters
    ----------
    src
        Predicted disparity map.
    tgt
        Input image (RGB or grayscale)
    """
    # Compute the Iverson bracket for adjacent pixels
    tgt_dx = torch.mean(torch.diff(tgt, dim=-1), dim=1, keepdim=True)
    tgt_dy = torch.mean(torch.diff(tgt, dim=-2), dim=1, keepdim=True)

    # Compute the partial disparity derivative
    src_dx = torch.diff(src, dim=-1)
    src_dy = torch.diff(src, dim=-2)

    loss_x = torch.mean(torch.mul(torch.abs(src_dx), torch.exp(-torch.abs(tgt_dx))))
    loss_y = torch.mean(torch.mul(torch.abs(src_dy), torch.exp(-torch.abs(tgt_dy))))
    return loss_x + loss_y


class IGSLoss(nn.Module):
    r"""Image-guided smoothness loss (IGS) loss."""

    def __init__(self) -> None:
        super().__init__()

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(self, src: Tensor, tgt: Tensor) -> Tensor:
        src = src.float()
        tgt = tgt.float()
        return compute_image_guided_smoothness_loss(src, tgt)


def gradient_loss(src: Tensor, tgt: Tensor, mask: Tensor) -> Tensor:
    n = mask.float().sum(dim=(1, 2))

    diff = src - tgt
    diff = torch.mul(mask, diff)

    grad_x = (diff[:, :, 1:] - diff[:, :, :-1]).abs()
    mask_x = torch.mul(mask[:, :, 1:], mask[:, :, :-1])
    grad_x = torch.mul(mask_x, grad_x)

    grad_y = (diff[:, 1:, :] - diff[:, :-1, :]).abs()
    mask_y = torch.mul(mask[:, 1:, :], mask[:, :-1, :])
    grad_y = torch.mul(mask_y, grad_y)

    image_loss = torch.sum(grad_x, (1, 2)) + torch.sum(grad_y, (1, 2))

    return _reduce_batch(image_loss, n)


class GradientLoss(nn.Module):
    def __init__(self, levels=4) -> None:
        super().__init__()

        self.levels = levels

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(self, src: Tensor, tgt: Tensor, mask: Tensor) -> Tensor:
        src = src.float()
        tgt = tgt.float()
        mask = mask.bool()

        total = []

        for scale in range(self.levels):
            step = pow(2, scale)

            total.append(
                gradient_loss(
                    src[:, ::step, ::step],
                    tgt[:, ::step, ::step],
                    mask[:, ::step, ::step],
                )
            )

        return torch.stack(total).sum()


def affine_loss(src: Tensor, tgt: Tensor, mask: Tensor) -> tuple[Tensor, Tensor]:
    # Let Ax = b be the system of equations to solve

    # A = [[a_00, a_01], [a_10, a_11]]
    a_00 = torch.sum(mask * src * src, (1, 2))
    a_01 = torch.sum(mask * src, (1, 2))
    a_11 = torch.sum(mask, (1, 2))

    # b = [b_0, b_1]
    b_0 = torch.sum(mask * src * tgt, (1, 2))
    b_1 = torch.sum(mask * tgt, (1, 2))

    # solution: x = A^-1 . b = [[a_11, -a_01], [-a_10, a_00]] / (a_00 * a_11 - a_01 * a_10) . b
    x_0 = torch.zeros_like(b_0)
    x_1 = torch.zeros_like(b_1)

    # Compute the determinant
    det = a_00 * a_11 - a_01 * a_01
    valid = det.nonzero()

    # Fill the solution
    x_0[valid] = (a_11[valid] * b_0[valid] - a_01[valid] * b_1[valid]) / det[valid]
    x_1[valid] = (-a_01[valid] * b_0[valid] + a_00[valid] * b_1[valid]) / det[valid]

    return x_0, x_1


class SSILoss(ScaledLossMixin, nn.Module):
    """Scale and Shift Invariant (SSI) loss with an additional Gradient loss mixed in."""

    def __init__(self, *, alpha=0.5, levels=4, **kwargs) -> None:
        super().__init__(**kwargs)

        self.rmse = RMSELoss()
        self.gradient = GradientLoss(levels=levels)
        self.alpha = alpha

    @T.override
    @torch.autocast("cuda", enabled=False)
    def forward(self, src: Tensor, tgt: Tensor, mask: Tensor) -> Tensor:
        src = src.float()
        tgt = tgt.float()
        mask = mask.bool()

        scale, shift = affine_loss(src, tgt, mask)
        ssi = scale.view(-1, 1, 1) * src + shift.view(-1, 1, 1)

        total = self.rmse(ssi, tgt, mask)
        if self.alpha > 0:
            total += self.alpha * self.gradient(ssi, tgt, mask)
        return total * self.scale



from __future__ import annotations

import torch
from torch import Tensor

from unipercept.utils.tensorclass import Tensorclass

from .transform import Rotate, Transform3d, TransformType, Translate

__all__ = ["Cameras", "get_world_to_view_transform"]


class Cameras(Tensorclass):
    """
    Parameters
    ----------
    K
        A calibration matrix of shape (N, 4, 4)
    R
        Rotation matrix of shape (N, 3, 3)
    T
        Translation matrix of shape (N, 3)
    image_size
        A tensor (int32) of shape (N, 2).
    """

    K: Tensor
    R: Tensor
    T: Tensor
    image_size: Tensor

    def get_camera_center(self) -> Tensor:
        w2v_trans = self.get_world_to_view_transform()
        assert isinstance(w2v_trans, Transform3d)

        P = w2v_trans.inverse(True).get_matrix()
        # the camera center is the translation component (the first 3 elements
        # of the last row) of the inverted world-to-view
        # transform (4x4 RT matrix)

        return P[:, 3, :3]

    def get_world_to_view_transform(self) -> Transform3d:
        return get_world_to_view_transform(R=self.R, T=self.T)

    def get_full_projection_transform(self) -> Transform3d:
        w2v = self.get_world_to_view_transform()
        v2p = self.get_projection_transform()
        return w2v.compose([v2p])

    def transform_points(self, points, eps: float | None = None) -> Tensor:
        world_to_proj_transform = self.get_full_projection_transform()
        return world_to_proj_transform.transform_points(points, eps=eps)

    def get_projection_transform(self) -> Transform3d:
        return Transform3d(
            self.K.transpose(1, 2).contiguous(), tt=TransformType.DEFAULT
        )

    def unproject_points(
        self,
        xy_depth: Tensor,
        world_coordinates=True,
    ) -> Tensor:
        if world_coordinates:
            to_camera_transform = self.get_full_projection_transform()
        else:
            to_camera_transform = self.get_projection_transform()

        assert isinstance(to_camera_transform, Transform3d)

        unprojection_transform = to_camera_transform.inverse(True)
        xy_inv_depth = torch.cat((xy_depth[..., :2], 1.0 / xy_depth[..., 2:3]), dim=-1)

        assert isinstance(xy_inv_depth, torch.Tensor)
        assert isinstance(unprojection_transform, Transform3d)

        return unprojection_transform.transform_points(xy_inv_depth)

    def get_principal_point(self) -> Tensor:
        proj_mat = self.get_projection_transform().get_matrix()
        return proj_mat[:, 2, :2]


def get_world_to_view_transform(R: Tensor, T: Tensor) -> Transform3d:
    """
    This function returns a Transform3d representing the transformation
    matrix to go from world space to view space by applying a rotation and
    a translation.

    PyTorch3D uses the same convention as Hartley & Zisserman.
    I.e., for camera extrinsic parameters R (rotation) and T (translation),
    we map a 3D point `X_world` in world coordinates to
    a point `X_cam` in camera coordinates with:
    `X_cam = X_world R + T`

    Parameters
    ----------
    R
        (N, 3, 3) matrix representing the rotation.
    T
        (N, 3) matrix representing the translation.

    Returns
    -------
        Transform3d object which represents the composed RT transformation.
    """
    # TODO: also support the case where RT is specified as one matrix
    # of shape (N, 4, 4).

    if T.shape[0] != R.shape[0]:
        msg = (
            f"Expected R, T to have the same batch dimension; got {R.shape[0]} "
            f", {T.shape[0]}"
        )
        raise ValueError(msg)
    if T.dim() != 2 or T.shape[1:] != (3,):
        msg = f"Expected T to have shape (N, 3); got {T.shape}"
        raise ValueError(msg)
    if R.dim() != 3 or R.shape[1:] != (3, 3):
        msg = f"Expected R to have shape (N, 3, 3); got {R.shape}"
        raise ValueError(msg)

    # Create a Transform3d object
    T_ = Translate(T)
    R_ = Rotate(R)

    return R_.compose([T_])



"""
Conditionally Parameterized Convolution.

Paper: https://arxiv.org/abs/1904.04971
"""

from __future__ import annotations

import math
from functools import partial
from typing import override

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

from unipercept.nn.conv.utils import NormActivationMixin
from unipercept.utils.function import to_2tuple

__all__ = ["CondConv2d"]


def get_condconv_initializer(initializer, num_experts, expert_shape):
    def condconv_initializer(weight) -> None:
        """CondConv initializer function."""
        num_params = np.prod(expert_shape)
        if (
            len(weight.shape) != 2
            or weight.shape[0] != num_experts
            or weight.shape[1] != num_params
        ):
            msg = "CondConv variables must have shape [num_experts, num_params]"
            raise (ValueError(msg))
        for i in range(num_experts):
            initializer(weight[i].view(expert_shape))

    return condconv_initializer


class CondConv2d(NormActivationMixin, nn.Module):
    r"""Conditional convolution layer."""

    __constants__ = ["in_channels", "out_channels"]

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        stride=1,
        padding: int | tuple[int, int] = 0,
        dilation=1,
        groups=1,
        bias=False,
        num_experts=4,
    ) -> None:
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = to_2tuple(kernel_size)
        self.stride = to_2tuple(stride)
        self.padding = to_2tuple(padding)
        self.dilation = to_2tuple(dilation)
        self.groups = groups
        self.num_experts = num_experts

        self.weight_shape = (
            self.out_channels,
            self.in_channels // self.groups,
            *self.kernel_size,
        )
        weight_num_param = 1
        for wd in self.weight_shape:
            weight_num_param *= wd
        self.weight = torch.nn.Parameter(
            torch.torch.Tensor(self.num_experts, weight_num_param)
        )

        if bias:
            self.bias_shape = (self.out_channels,)
            self.bias = torch.nn.Parameter(
                torch.torch.Tensor(self.num_experts, self.out_channels)
            )
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        init_weight = get_condconv_initializer(
            partial(nn.init.kaiming_uniform_, a=math.sqrt(5)),
            self.num_experts,
            self.weight_shape,
        )
        init_weight(self.weight)
        if self.bias is not None:
            fan_in = np.prod(self.weight_shape[1:])
            bound = 1 / math.sqrt(fan_in)
            init_bias = get_condconv_initializer(
                partial(nn.init.uniform_, a=-bound, b=bound),
                self.num_experts,
                self.bias_shape,
            )
            init_bias(self.bias)

    @override
    def forward(self, x, routing_weights):
        B, C, H, W = x.shape
        weight = torch.matmul(routing_weights, self.weight)
        new_weight_shape = (
            B * self.out_channels,
            self.in_channels // self.groups,
            *self.kernel_size,
        )
        weight = weight.view(new_weight_shape)
        bias = None
        if self.bias is not None:
            bias = torch.matmul(routing_weights, self.bias)
            bias = bias.view(B * self.out_channels)
        # move batch elements with channels so each batch element can be efficiently convolved with separate kernel
        # reshape instead of view to work with channels_last input
        x = x.reshape(1, B * C, H, W)
        x = F.conv2d(
            x,
            weight,
            bias,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            groups=self.groups * B,
        )
        return x.permute([1, 0, 2, 3]).view(
            B, self.out_channels, x.shape[-2], x.shape[-1]
        )



import functools
import math
import typing as T

import torch
from torch import nn

from unipercept.nn.activations import ActivationSpec, get_activation
from unipercept.nn.init import InitMixin, InitMode, InitSpec
from unipercept.nn.norms import NormSpec, get_default_bias, get_norm
from unipercept.types import Tensor

from .utils import NormActivationSupport

__all__ = [
    "Conv2d",
    "ConvTranspose2d",
    "IntParam1d",
    "IntParam2d",
    "IntParam3d",
    "Linear2d",
    "PaddingMode",
    "PaddingParam",
    "Separable2d",
    "Standard2d",
]

type IntParam1d = int | tuple[int] | T.Sequence[int]
type IntParam2d = int | tuple[int, int] | T.Sequence[int]
type IntParam3d = int | tuple[int, int, int] | T.Sequence[int]
type PaddingParam = T.Literal["valid", "same"] | int
type PaddingMode = T.Literal["zeros", "circular", "reflect"]


class Conv2d(NormActivationSupport, InitMixin, nn.Conv2d):
    r"""
    A 2D convolutional layer with optional normalization and activation.

    Notes
    -----
    Current benchmarks show that this is the fastest way to implement a convolutional
    layer with added normalization and activation, as it avoids the overhead of
    using a `Sequential` container. This may become redundant in the future as the
    PyTorch framework evolves.
    """

    norm: nn.Module | None
    activation: nn.Module | None

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: IntParam2d,
        *,
        bias: bool | None = None,
        norm: NormSpec | None = None,
        activation: ActivationSpec | None = None,
        **kwargs,
    ) -> None:
        norm_layer = get_norm(norm, out_channels)
        act_layer = get_activation(activation)
        bias = get_default_bias(bias, norm_layer)

        super().__init__(in_channels, out_channels, kernel_size, bias=bias, **kwargs)

        if norm_layer is not None:
            self.norm = norm_layer
        else:
            self.register_module("norm", None)
        if act_layer is not None:
            self.activation = act_layer
        else:
            self.register_module("activation", None)

        self.reset_parameters()

    @T.override
    def forward(self, input: Tensor) -> Tensor:
        result = self._conv_forward(input, self.weight, self.bias)
        if self.norm is not None:
            result = self.norm(result)
        if self.activation is not None:
            result = self.activation(result)
        return result


class Linear2d(Conv2d):
    r"""
    A convolutional layer with a kernel size of 1. This is equivalent to a linear layer,
    but with spatial dimensions.

    This is a convenience class that constraints the initialization to not have
    any parameters that would make it inconsistent with its intended purpose of being
    a linear mapping (kernel size, padding, etc.).
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int | None = None,
        *,
        bias: bool | None = None,
        norm: NormSpec | None = None,
        activation: ActivationSpec | None = None,
        init_mode: InitMode | None = None,
        init_spec: InitSpec | None = None,
    ) -> None:
        if out_channels is None:
            out_channels = in_channels
        super().__init__(
            in_channels,
            out_channels,
            1,
            padding=0,
            stride=1,
            dilation=1,
            groups=1,
            bias=bias,
            norm=norm,
            activation=activation,
            init_mode=init_mode,
            init_spec=init_spec,
        )


class ConvTranspose2d(nn.ConvTranspose2d):
    norm: nn.Module | None
    activation: nn.Module | None

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *args,
        bias: bool | None = None,
        norm: NormSpec | None = None,
        activation: ActivationSpec | None = None,
        **kwargs,
    ) -> None:
        norm_layer = get_norm(norm, out_channels)
        act_layer = get_activation(activation)
        bias = get_default_bias(bias, norm_layer)

        super().__init__(in_channels, out_channels, *args, bias=bias, **kwargs)

        if norm_layer is not None:
            self.norm = norm_layer
        else:
            self.register_module("norm", None)
        if act_layer is not None:
            self.activation = act_layer
        else:
            self.register_module("activation", None)

        self.reset_parameters()

    @T.override
    def forward(self, input: Tensor, output_size: list[int] | None = None) -> Tensor:
        result = self._transpose_conv_forward(input, output_size)
        if self.norm is not None:
            result = self.norm(result)
        if self.activation is not None:
            result = self.activation(result)
        return result

    def _transpose_conv_forward(
        self, input: Tensor, output_size: list[int] | None
    ) -> Tensor:
        r"""Copied from `torch.nn.modules.conv.ConvTranspose2d` to allow for custom forwards."""
        if self.padding_mode != "zeros":
            msg = "Only `zeros` padding mode is supported for ConvTranspose2d"
            raise ValueError(msg)

        assert isinstance(self.padding, tuple)
        # One cannot replace List by Tuple or Sequence in "_output_padding" because
        # TorchScript does not support `Sequence[T]` or `Tuple[T, ...]`.
        num_spatial_dims = 2
        output_padding = self._output_padding(
            input,
            output_size,
            self.stride,
            self.padding,
            self.kernel_size,  # type: ignore[arg-type]
            num_spatial_dims,
            self.dilation,
        )  # type: ignore[arg-type]

        return nn.functional.conv_transpose2d(
            input,
            self.weight,
            self.bias,
            self.stride,
            self.padding,
            output_padding,
            self.groups,
            self.dilation,
        )

    @T.override
    def reset_parameters(self) -> None:
        weight = self.weight.data
        feature = math.ceil(weight.size(2) / 2)
        channels = (2 * feature - 1 - feature % 2) / (2.0 * feature)
        for i in range(weight.size(2)):
            for j in range(weight.size(3)):
                weight[0, 0, i, j] = (1 - math.fabs(i / feature - channels)) * (
                    1 - math.fabs(j / feature - channels)
                )
        for channels in range(1, weight.size(0)):
            weight[channels, 0, :, :] = weight[0, 0, :, :]


class Standard2d(Conv2d):
    """
    Implements weight standardization with learnable gain.
    Paper: https://arxiv.org/abs/2101.08692.

    Note that this layer must *always* be followed by some form of normalization.
    """

    scale: T.Final[float]

    def __init__(self, *args, gamma=1.0, eps=1e-6, gain=1.0, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.gain = nn.Parameter(torch.full((self.out_channels, 1, 1, 1), gain))
        self.scale = float(gamma * self.weight[0].numel() ** -0.5)
        self.eps = eps

    @T.override
    def forward(self, input: Tensor) -> Tensor:
        weight = nn.functional.batch_norm(
            self.weight.reshape(1, self.out_channels, -1),
            None,
            None,
            weight=(self.gain * self.scale).view(-1),
            training=True,
            momentum=0.0,
            eps=self.eps,
        ).reshape_as(self.weight)
        return self._conv_forward(input, weight, self.bias)


class Separable2d(NormActivationSupport, InitMixin, nn.Module):
    """
    Impements a depthwise-seperable 2d convolution. This is a combination of a depthwise convolution and a pointwise
    convolution. The depthwise convolution is a convolution with a kernel size of 1x1xKxK and a pointwise convolution
    is a convolution with a kernel size of 1x1x1x1. The depthwise convolution is applied to each input channel
    individually and the pointwise convolution is applied to the output of the depthwise convolution. This is
    equivalent to a convolution with a kernel size of 1x1xKxK.
    """

    in_channels: T.Final[int]
    hidden_channels: T.Final[int]
    out_channels: T.Final[int]
    reverse: T.Final[bool]

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: IntParam2d,
        *,
        stride: IntParam2d = 1,
        padding: IntParam2d | str = "valid",
        padding_mode: str = "zeros",
        dilation: IntParam2d = 1,
        groups: int | None = None,
        expansion: int = 1,
        bias: bool | None = None,
        norm: NormSpec | None = None,
        activation: ActivationSpec | None = None,
        init_mode: InitMode | None = None,
        init_spec: InitSpec | None = None,
        reverse: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        in_channels
            Number of input channels.
        out_channels
            Number of output channels.
        expansion
            Expansion factor for the middle channels.
        bias
            Whether to add a bias to the *pointwise* convolution.
        groups
            Number of groups for the *depthwise* convolution.
            If a float is provided, it is interpreted as a fraction
            of the input channels.
            If None (default) this is equal to the number of input channels.
        init
            Initialization method for the weights.
        **kwargs
            Additional arguments passed to the *depthwise* convolution.
        """
        hidden_channels = int(in_channels * expansion)
        if hidden_channels % in_channels != 0:
            msg = f"Number of input channels ({in_channels}) must be a divisor of the middle channels ({hidden_channels})"
            raise ValueError(msg)
        if groups is None:
            groups = in_channels
        if isinstance(groups, float):
            assert groups < 1.0
            groups = int(groups * in_channels)

        self.in_channels = in_channels
        self.hidden_channels = hidden_channels
        self.out_channels = out_channels
        self.reverse = reverse

        if padding == "valid":
            padding = (
                [k // 2 for k in kernel_size]
                if isinstance(kernel_size, T.Iterable)
                else kernel_size // 2
            )

        depthwise = functools.partial(
            Conv2d,
            in_channels,
            hidden_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            padding_mode=padding_mode,
            dilation=dilation,
            bias=False,
            groups=groups,
            init_mode=init_mode,
            init_spec=init_spec,
        )
        pointwise = functools.partial(
            Linear2d,
            hidden_channels,
            out_channels,
            bias=bias,
            init_mode=init_mode,
            init_spec=init_spec,
        )

        super().__init__()

        if not self.reverse:
            self.depthwise = depthwise()
            self.pointwise = pointwise(
                norm=norm,
                activation=activation,
            )
        else:
            self.depthwise = pointwise()
            self.pointwise = depthwise(
                norm=norm,
                activation=activation,
            )
        self.reset_parameters()

    def _separable_conv(self, input: Tensor) -> Tensor:
        if self.reverse:
            result = self.pointwise(input)
            result = self.depthwise(result)
        else:
            result = self.depthwise(input)
            result = self.pointwise(result)
        return result

    @T.override
    def forward(self, input: Tensor) -> Tensor:
        return self._separable_conv(input)



"""
It is used to determine the file format of the images in the dataset.

This file is an extended version Python 3.11 standard library,
this was done because the module will be removed in Python 3.13.

Source: https://raw.githubusercontent.com/python/cpython/3.11/Lib/imghdr.py

Added featues are:
- enumerations for image formats.
- a function to read the image size from the file header.
"""

from __future__ import annotations

import struct
from enum import StrEnum
from enum import auto as A
from os import PathLike
from pathlib import Path
from typing import IO, NamedTuple

__all__ = ["Format", "Size", "size", "what"]


# ------------------------------------ #
# Determines the type of an image file #
# ------------------------------------ #


class Format(StrEnum):
    JPEG = A()
    PNG = A()
    GIF = A()
    TIFF = A()
    BMP = A()
    PBM = A()
    PGM = A()
    PPM = A()
    RAST = A()
    XBM = A()
    WEBP = A()
    EXR = A()
    RGB = A()


def what(file: str | PathLike | IO, h=None) -> Format | None:
    """Tries to determine the type of an image file given its file path."""
    f = None
    try:
        if h is None:
            if isinstance(file, (str, PathLike)):
                f = open(file, "rb")
                h = f.read(32)
            else:
                location = file.tell()
                h = file.read(32)
                file.seek(location)
        for tf in tests:
            res = tf(h, f)
            if res:
                return res
    finally:
        if f:
            f.close()
    return None


# ------------------------------- #
# Subroutines per image file type #
# ------------------------------- #

tests = []


def test_jpeg(h, f):
    """JPEG data with JFIF or Exif markers; and raw JPEG."""
    if h[6:10] in (b"JFIF", b"Exif") or h[:4] == b"\xff\xd8\xff\xdb":
        return Format.JPEG
    return None


tests.append(test_jpeg)


def test_png(h, f):
    if h.startswith(b"\211PNG\r\n\032\n"):
        return Format.PNG
    return None


tests.append(test_png)


def test_gif(h, f):
    """GIF ('87 and '89 variants)."""
    if h[:6] in (b"GIF87a", b"GIF89a"):
        return Format.GIF
    return None


tests.append(test_gif)


def test_tiff(h, f):
    """TIFF (can be in Motorola or Intel byte order)."""
    if h[:2] in (b"MM", b"II"):
        return Format.TIFF
    return None


tests.append(test_tiff)


def test_rgb(h, f):
    """SGI image library."""
    if h.startswith(b"\001\332"):
        return Format.RGB
    return None


tests.append(test_rgb)


def test_pbm(h, f):
    """PBM (portable bitmap)."""
    if len(h) >= 3 and h[0] == ord(b"P") and h[1] in b"14" and h[2] in b" \t\n\r":
        return Format.PBM
    return None


tests.append(test_pbm)


def test_pgm(h, f):
    """PGM (portable graymap)."""
    if len(h) >= 3 and h[0] == ord(b"P") and h[1] in b"25" and h[2] in b" \t\n\r":
        return Format.PGM
    return None


tests.append(test_pgm)


def test_ppm(h, f):
    """PPM (portable pixmap)."""
    if len(h) >= 3 and h[0] == ord(b"P") and h[1] in b"36" and h[2] in b" \t\n\r":
        return Format.PPM
    return None


tests.append(test_ppm)


def test_rast(h, f):
    """Sun raster file."""
    if h.startswith(b"\x59\xa6\x6a\x95"):
        return Format.RAST
    return None


tests.append(test_rast)


def test_xbm(h, f):
    """X bitmap (X10 or X11)."""
    if h.startswith(b"#define "):
        return Format.XBM
    return None


tests.append(test_xbm)


def test_bmp(h, f):
    if h.startswith(b"BM"):
        return Format.BMP
    return None


tests.append(test_bmp)


def test_webp(h, f):
    if h.startswith(b"RIFF") and h[8:12] == b"webp":
        return Format.WEBP
    return None


tests.append(test_webp)


def test_exr(h, f):
    if h.startswith(b"\x76\x2f\x31\x01"):
        return Format.EXR
    return None


tests.append(test_exr)


# -------------------------------- #
# Determine image size from header #
# -------------------------------- #


class Size(NamedTuple):
    """Image size, in pixels (width, height)."""

    width: int
    height: int


def size(path: Path | str) -> Size:
    """
    Read image size from the file headers; does not require file to be opened.
    This is useful in sitations where a fast assessment of image sizes must be
    made, like when providing metadata to a dataset record.

    Parameters
    ----------
    path
        Path to the image file.

    Returns
    -------
        Width and height of the image.
    """
    path = str(path)

    with open(path, "rb") as fh:
        head = fh.read(24)
        if len(head) != 24:
            msg = f"bad length read {len(head)} for {path}!"
            raise ValueError(msg)

        image_type = what(path)
        if image_type == "png":
            check = struct.unpack(">i", head[4:8])[0]
            if check != 0x0D0A1A0A:
                msg = f"Check failed: {check}!"
                raise ValueError(msg)
            width, height = struct.unpack(">ii", head[16:24])
        elif image_type == "gif":
            width, height = struct.unpack("<HH", head[6:10])
        elif image_type == "jpeg":
            try:
                fh.seek(0)  # Read 0xff next
                size = 2
                ftype = 0
                while not 0xC0 <= ftype <= 0xCF:
                    fh.seek(size, 1)
                    byte = fh.read(1)
                    while ord(byte) == 0xFF:
                        byte = fh.read(1)
                    ftype = ord(byte)
                    size = struct.unpack(">H", fh.read(2))[0] - 2
                # We are at a SOFn block
                fh.seek(1, 1)  # Skip `precision' byte.
                height, width = struct.unpack(">HH", fh.read(4))
            except Exception:  # IGNORE:W0703
                raise
        else:
            raise NotImplementedError(image_type)

    return Size(width=width, height=height)


# --------------------#
# Small test program #
# --------------------#


def test() -> None:
    import sys

    recursive = 0
    if sys.argv[1:] and sys.argv[1] == "-r":
        del sys.argv[1:2]
        recursive = 1
    try:
        if sys.argv[1:]:
            testall(sys.argv[1:], recursive, 1)
        else:
            testall(["."], recursive, 1)
    except KeyboardInterrupt:
        sys.stderr.write("\n[Interrupted]\n")
        sys.exit(1)


def testall(list, recursive, toplevel) -> None:
    import os
    import sys

    for filename in list:
        if os.path.isdir(filename):
            print(filename + "/:", end=" ")
            if recursive or toplevel:
                print("recursing down:")
                import glob

                names = glob.glob(os.path.join(glob.escape(filename), "*"))
                testall(names, recursive, 0)
            else:
                print("*** directory (use -r) ***")
        else:
            print(filename + ":", end=" ")
            sys.stdout.flush()
            try:
                print(what(filename))
            except OSError:
                print("*** not found ***")


if __name__ == "__main__":
    test()



"""Unified perception library."""

from unipercept.utils.python import defer_imports as _defer_imports

__version__: str
__all__ = ["log", "nn", "ops", "state", "tensors", "types", "utils", "vision"]

__getattr__, __dir__ = _defer_imports(__name__, __all__, distribution="unipercept")



import torch
from torch import nn

from unipercept.nn.transformer.attention import LinearProject


class GatedFeedForward(nn.Module):
    def __init__(
        self, dim: int, hidden_dim: int, proj: LinearProject = nn.Linear
    ) -> None:
        super().__init__()

        self.proj_g = proj(dim, hidden_dim, bias=False)
        self.proj_o = proj(hidden_dim, dim, bias=False)
        self.proj_i = proj(dim, hidden_dim, bias=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.proj_o(nn.functional.silu(self.proj_g(x)) * self.proj_i(x))  # type: ignore



r"""
Mask tensors.
============

Implements mask tensors for semantic and instance segmentation maps, and their
(panoptic segmentation) joint representation.
"""

import enum as E
import typing as T

import expath
import safetensors.torch as safetensors
import torch
import torchvision.io

from unipercept.tensors.helpers import write_png_l16, write_png_rgb
from unipercept.tensors.registry import pixel_maps
from unipercept.types import Tensor

from ._torchvision import MaskTensor

__all__ = [
    "PanopticFormat",
    "PanopticTensor",
    "PanopticTensorLike",
    "load_panoptic",
    "save_panoptic",
]

_L = T.TypeVar("_L", int, Tensor)
_BYTE_OFFSET: T.Final = 256


class PanopticFormat(E.StrEnum):
    """
    Enumerates the different formats of labels that are supported. Uses the name of
    the dataset that introduced the format.
    """

    CITYSCAPES = E.auto()
    CITYSCAPES_VPS = E.auto()
    CITYSCAPES_DVPS = E.auto()
    KITTI = E.auto()
    VISTAS = E.auto()
    WILD_DASH = E.auto()
    TORCH = E.auto()
    SAFETENSORS = E.auto()
    PNG_UINT16 = E.auto()
    PNG_L16 = E.auto()


@pixel_maps.register()
class PanopticTensor(MaskTensor):
    """
    Implements a panoptic segmentation map, where each pixel has the value:
        category_id * divisor + instance_id.
    """

    DIVISOR: T.ClassVar[int] = 2**15  # same for all datasets
    IGNORE: T.ClassVar[int] = -1

    @classmethod
    @torch.no_grad()
    def read(cls, path: expath.PathType, **meta_kwds) -> T.Self:
        return load_panoptic(path, **meta_kwds).as_subclass(cls)

    def save(
        self,
        path: expath.PathType,
        *,
        format: PanopticFormat | str | None = None,
        **meta_kwds,
    ) -> None:
        return save_panoptic(self, path, format=format, **meta_kwds)

    @classmethod
    def default(cls, shape: torch.Size, device: torch.device | str = "cpu") -> T.Self:
        return torch.full(
            shape, cls.IGNORE * cls.DIVISOR, dtype=torch.long, device=device
        ).as_subclass(cls)

    @classmethod
    def default_like(cls, other: Tensor) -> T.Self:
        """Returns a default instance of this class with the same shape as the given tensor."""
        return cls(
            torch.full_like(
                other, fill_value=cls.IGNORE * cls.DIVISOR, dtype=torch.long
            )
        )

    @classmethod
    @T.override
    def wrap_like(
        cls,
        other: T.Self,
        tensor: Tensor,
    ) -> T.Self:
        return tensor.to(dtype=torch.long, non_blocking=True).as_subclass(cls)

    @classmethod
    def from_semantic(cls, semantic: Tensor) -> T.Self:
        """Create an instance from a semantic segmentation map by setting the instance IDs to 0."""
        if not torch.compiler.is_compiling():
            cls.must_be_semantic_map(semantic)
        return (semantic * cls.DIVISOR).as_subclass(PanopticTensor)

    @classmethod
    def from_parts(cls, semantic: Tensor, instance: Tensor) -> T.Self:
        """
        Create an instance from a semantic segmentation and instance segmentation map by combining them
        using the global ``divisor``.
        """
        if not torch.compiler.is_compiling():
            if semantic.shape != instance.shape:
                msg = f"Expected tensors of the same shape, got {semantic.shape} and {instance.shape}"
                raise ValueError(msg)
            cls.must_be_semantic_map(semantic)
            cls.must_be_instance_map(instance)

        semantic = semantic.to(dtype=torch.long, non_blocking=True)
        instance = instance.to(dtype=torch.long, non_blocking=True)

        ignore_mask = semantic == cls.IGNORE
        panoptic = instance + semantic * cls.DIVISOR
        panoptic[ignore_mask] = cls.IGNORE

        return panoptic.as_subclass(PanopticTensor)

    @classmethod
    def from_combined(cls, encoded_map: Tensor | T.Any, divisor: int) -> T.Self:
        """
        Decompose an encoded map into a semantic segmentation and instance segmentation map, then combine
        again using the global ``divisor``.
        """
        encoded_map = torch.as_tensor(encoded_map)
        assert encoded_map.dtype in (torch.int32, torch.int64), encoded_map.dtype

        sem_id = torch.floor_divide(encoded_map, divisor)
        ins_id = torch.remainder(encoded_map, divisor)
        ins_id = torch.where(encoded_map >= 0, ins_id, 0)

        return PanopticTensor.from_parts(sem_id, ins_id)

    @T.overload
    def to_parts(self: Tensor) -> Tensor: ...

    @T.overload
    def to_parts(self: Tensor, as_tuple=True) -> tuple[Tensor, Tensor]: ...

    def to_parts(
        self: Tensor, as_tuple: bool = False
    ) -> Tensor | tuple[Tensor, Tensor]:
        """
        Split the semantic and instance segmentation maps, returing a tensor of size [..., 2].
        The first channel contains the semantic segmentation map, the second channel contains the instance
        id that is NOT UNIQUE for each class.
        """
        ignore_mask = self == PanopticTensor.IGNORE
        sem = torch.floor_divide(self, PanopticTensor.DIVISOR)
        ins = torch.remainder(self, PanopticTensor.DIVISOR)
        ins[ignore_mask] = 0
        if as_tuple:
            return sem.as_subclass(MaskTensor), ins.as_subclass(MaskTensor)
        return torch.stack((sem, ins), dim=-1).as_subclass(MaskTensor)

    @staticmethod
    def parse_label(label: _L) -> tuple[_L, _L]:
        """Parse a label into a semantic and instance ID."""
        sem_id: _L
        ins_id: _L

        if isinstance(label, int):
            sem_id = label // PanopticTensor.DIVISOR
            ins_id = label % PanopticTensor.DIVISOR if label >= 0 else 0
        else:
            sem_id = torch.floor_divide(label, PanopticTensor.DIVISOR)
            ins_id = torch.remainder(label, PanopticTensor.DIVISOR)
            ins_id = torch.where(label >= 0, ins_id, 0)

        return sem_id, ins_id

    @classmethod
    @T.overload
    def is_void(cls, label: int) -> bool: ...

    @classmethod
    @T.overload
    def is_void(cls, label: Tensor) -> Tensor: ...

    @classmethod
    def is_void(cls, label: Tensor | int) -> Tensor | bool:
        return label < 0

    def get_semantic_map(self: Tensor) -> MaskTensor:
        return torch.floor_divide(self, PanopticTensor.DIVISOR).as_subclass(MaskTensor)

    def get_semantic_masks(self: Tensor) -> T.Iterator[tuple[int, MaskTensor]]:
        """Return a list of masks, one for each semantic class."""
        sem_map = PanopticTensor.get_semantic_map(self)
        uq = torch.unique(sem_map)
        yield from (
            (int(u), (sem_map == u).as_subclass(MaskTensor))
            for u in uq
            if u != PanopticTensor.IGNORE
        )

    def get_semantic_mask(self, class_id: int) -> MaskTensor:
        """Return a mask for the specified semantic class."""
        return (self.get_semantic_map() == class_id).as_subclass(MaskTensor)

    def unique_semantics(self) -> Tensor:
        """Count the number of unique semantic classes."""
        uq = torch.unique(self.get_semantic_map())
        return uq[uq >= 0]

    def get_instance_map(self: Tensor, return_labels: bool = True) -> MaskTensor:
        # old: does not support same sub-id for different classes
        ins_ids = torch.remainder(self, PanopticTensor.DIVISOR)
        if return_labels:
            mask = torch.where((ins_ids > 0) & (self != PanopticTensor.IGNORE), self, 0)
        else:
            mask = torch.where(self != PanopticTensor.IGNORE, ins_ids, 0)
        if not torch.compiler.is_compiling():
            mask = mask.as_subclass(MaskTensor)
        return mask

    def get_instance_masks(self: Tensor) -> T.Iterator[tuple[int, MaskTensor]]:
        """Return a list of masks, one for each instance."""
        ins_map = PanopticTensor.get_instance_map(self)
        uq = torch.unique(ins_map)
        yield from (
            (int(u), (ins_map == u).as_subclass(MaskTensor)) for u in uq if u > 0
        )

    def get_instance_mask(self: Tensor, instance_id: int) -> MaskTensor:
        """Return a mask for the specified instance."""
        return (PanopticTensor.get_instance_map(self) == instance_id).as_subclass(
            MaskTensor
        )

    def get_masks_by_label(
        self: Tensor, *, with_void: bool = False, as_tensor: bool = False
    ) -> T.Iterator[tuple[Tensor, MaskTensor]]:
        """
        Iterate pairs of labels and masks, where each masks corresponds to a unique
        label.
        """
        for pan_id in self.unique():
            if not as_tensor:
                pan_id = pan_id.detach().item()
            if PanopticTensor.is_void(pan_id) and not with_void:
                continue
            yield pan_id, (self == pan_id).as_subclass(MaskTensor)

    def get_masks(
        self: Tensor, **kwargs
    ) -> T.Iterator[tuple[Tensor, Tensor, MaskTensor]]:
        """Return a mask for each semantic class and instance (if any)."""
        for pan_id, mask in PanopticTensor.get_masks_by_label(self, **kwargs):
            sem_id, ins_id = PanopticTensor.parse_label(pan_id)
            yield sem_id, ins_id, mask

    def unique_instances(self: Tensor) -> Tensor:
        """Count the number of unique instances for each semantic class."""
        ins_mask = PanopticTensor.get_instance_map(self) != PanopticTensor.IGNORE
        return torch.unique(self[ins_mask])

    def remove_instances_(self: Tensor, semantic_list: T.Iterable[int]) -> None:
        """Remove instances for the specified semantic classes."""
        sem_map, ins_map = PanopticTensor.to_parts(self, as_tuple=True)

        # Compute candidate map where all pixels that are not in the semantic list are set to -1
        can_map = torch.where(ins_map > 0, sem_map, PanopticTensor.IGNORE)

        # Set all pixels that are not in the semantic list to 0
        for class_ in semantic_list:
            self[can_map == class_] = class_ * PanopticTensor.DIVISOR

    def remove_instances(self: Tensor, semantic_list: T.Iterable[int]) -> Tensor:
        sem_map, _ = PanopticTensor.to_parts(self, as_tuple=True)
        replace = torch.tensor(
            list(semantic_list), dtype=sem_map.dtype, device=sem_map.device
        )
        return self.where(
            ~torch.isin(sem_map, replace), sem_map * PanopticTensor.DIVISOR
        )

    def translate_semantic_(
        self: Tensor,
        translation: T.Mapping[int, int],
        inverse: bool = False,
        *,
        raises: bool = False,
    ) -> Tensor:
        """
        DEPRECATED.

        Use `recode_all_` instead.
        """
        sem_map, ins_map = PanopticTensor.to_parts(self, as_tuple=True)
        self.fill_(PanopticTensor.IGNORE)

        if raises:
            self_ids = set(map(int, torch.unique(sem_map.detach().cpu()).tolist()))
            keys_ids = set(map(int, translation.keys()))
            miss_ids = self_ids - keys_ids

            if len(miss_ids) > 0:
                msg = f"Missing {miss_ids} in translations {translation}"
                raise ValueError(msg)

        for (
            old_id,
            new_id,
        ) in translation.items():
            if inverse:
                old_id, new_id = new_id, old_id
            mask = sem_map == old_id
            self[mask] = new_id * PanopticTensor.DIVISOR + ins_map[mask]

        return self

    def recode(
        self: Tensor,
        old_cat: int,
        new_cat: int,
        *,
        missing_ok: bool = False,
        inplace=False,
    ) -> Tensor:
        sem_map, ins_map = PanopticTensor.to_parts(self, as_tuple=True)
        mask = sem_map == old_cat

        # Check if there are any pixels for the given semantic category
        if not missing_ok and not mask.any():
            msg = f"No pixels found for semantic category {old_cat}"
            raise ValueError(msg)

        # Inplace variant updates entries at the given mask
        if inplace:
            self[mask] = new_cat * PanopticTensor.DIVISOR + ins_map[mask]
            return self

        # Non-inplace variant uses `Tensor.where(mask, otherwise)` to update entries
        # in the semantic map, and then creates a new panoptic map
        sem_map = sem_map.where(~mask, new_cat)
        return PanopticTensor.from_parts(sem_map, ins_map)

    def recode_(self: Tensor, *translation, **kwargs) -> Tensor:
        return PanopticTensor.recode(self, *translation, inplace=True, **kwargs)

    def recode_all(
        self: Tensor,
        pairs: T.Mapping[int, int],
        *,
        reverse: bool = False,
        fill: int = IGNORE,
        inplace: bool = False,
    ) -> Tensor:
        """Recode the semantic classes of the map using the given translation mapping."""
        # Reverse the mapping of `old_id` -> `new_id` to `new_id` -> `old_id`
        if reverse:
            pairs = {v: k for k, v in pairs.items()}

        # Split the panoptic map into semantic and instance parts
        sem_map, ins_map = PanopticTensor.to_parts(self, as_tuple=True)

        # Convert the fill value to a PanopticID
        fill = fill * PanopticTensor.DIVISOR

        # Result panoptic map
        if inplace:
            result = self
            result.fill_(fill)
        else:
            result = torch.full_like(self, fill, dtype=self.dtype)

        # Recode each semantic class
        for old_cat, new_cat in pairs.items():
            new_id = new_cat * PanopticTensor.DIVISOR
            if inplace:
                mask = sem_map == old_cat
                result[mask] = new_id + ins_map[mask]
            else:
                result = result.where(sem_map != old_cat, new_id + ins_map)
        return result

    def recode_all_(
        self: Tensor,
        *args,
        **kwargs,
    ) -> Tensor:
        r"""In-place version of :meth:`recode_all`."""
        return PanopticTensor.recode_all(self, *args, inplace=True, **kwargs)

    def get_nonempty(self: Tensor) -> MaskTensor:
        """Return a new instance with only the non-empty pixels."""
        return self[self >= 0].as_subclass(MaskTensor)

    @classmethod
    def must_be_semantic_map(cls, t: Tensor) -> None:
        if t.ndim < 2:
            msg = f"Expected 2D tensor, got {t.ndim}D tensor"
        elif t.dtype not in (torch.int32, torch.int64):
            msg = f"Expected int32 or int64 tensor, got {t.dtype}"
        elif (t_min := t.min()) < cls.IGNORE:
            msg = f"Expected non-negative values other than {cls.IGNORE}, got {t_min.item()}"
        elif (t_max := t.max()) >= cls.DIVISOR:
            msg = f"Expected values < {cls.DIVISOR}, got {t_max.item()}"
        else:
            return
        raise ValueError(msg)

    @classmethod
    def is_semantic_map(cls, t: Tensor) -> bool:
        try:
            cls.must_be_semantic_map(t)
        except ValueError:
            return False
        return True

    @classmethod
    def must_be_instance_map(cls, t: Tensor) -> None:
        if t.ndim < 2:
            msg = f"Expected 2D tensor, got {t.ndim}D tensor"
        elif t.dtype not in (torch.int32, torch.int64):
            msg = f"Expected int32 or int64 tensor, got {t.dtype}"
        elif (t_min := t.min()) < cls.IGNORE:
            msg = f"Expected non-negative values, got {t_min.item()}"
        elif (t_max := t.max()) >= cls.DIVISOR:
            msg = f"Expected values < {cls.DIVISOR}, got {t_max.item()}"
        else:
            return
        raise ValueError(msg)

    @classmethod
    def is_instance_map(cls, t: Tensor) -> bool:
        try:
            cls.must_be_instance_map(t)
        except ValueError:
            return False
        return True


type PanopticTensorLike = PanopticTensor | Tensor


def load_panoptic(input: expath.PathType | Tensor, /, **meta_kwds) -> PanopticTensor:
    """Read a panoptic map from a file."""
    from .helpers import get_kwd

    data_format = get_kwd(meta_kwds, "format", PanopticFormat)

    match data_format:
        case PanopticFormat.SAFETENSORS:
            if isinstance(input, torch.Tensor):
                msg = "Expected a path to a file, not a tensor."
                raise TypeError(msg)
            input = str(expath.locate(input))
            labels = safetensors.load_file(input)["data"].as_subclass(PanopticTensor)
        case PanopticFormat.TORCH:
            if isinstance(input, torch.Tensor):
                msg = "Expected a path to a file, not a tensor."
                raise TypeError(msg)
            with expath.open(input, "rb") as fh:
                labels = torch.load(fh, map_location=torch.device("cpu"))
            labels = labels.as_subclass(PanopticTensor)
            assert labels is not None
            assert isinstance(labels, (PanopticTensor, torch.Tensor)), type(labels)
        case PanopticFormat.CITYSCAPES:
            divisor = 1000
            void_id = 255
            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            img = torchvision.io.decode_image(
                input, mode=torchvision.io.ImageReadMode.UNCHANGED
            )
            assert img.ndim == 3, f"Expected 3D tensor, got {img.ndim}D tensor"

            map_ = (
                img[:, :, 0]
                + _BYTE_OFFSET * img[:, :, 1]
                + _BYTE_OFFSET * _BYTE_OFFSET * img[:, :, 2]
            )
            map_ = torch.where(map_ > 0, map_, void_id)
            map_ = torch.where(map_ < divisor, map_ * divisor, map_ + 1)

            labels = PanopticTensor.from_combined(map_, divisor)
        case PanopticFormat.CITYSCAPES_VPS:
            divisor = 1000
            void_id = 255

            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            img = torchvision.io.decode_image(
                input, mode=torchvision.io.ImageReadMode.UNCHANGED
            )
            assert img.dtype == torch.uint16
            assert img.ndim == 3, (
                f"Expected 3D tensor, got {img.ndim}D tensor {img.shape}"
            )

            # Convert to `int` while torch support for `uint16` is not available
            img = img.int()
            has_instance = img >= divisor

            ids = torch.where(has_instance, (img % divisor) + 1, 0)
            sem = torch.where(has_instance, img // divisor, img)
            sem[sem == void_id] = -1

            labels = PanopticTensor.from_parts(sem, ids)
        case PanopticFormat.CITYSCAPES_DVPS:
            divisor = 1000
            void_id = 32

            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            img = torchvision.io.decode_image(
                input, mode=torchvision.io.ImageReadMode.UNCHANGED
            )
            assert img.ndim == 3, (
                f"Expected 3D tensor, got {img.ndim}D tensor {img.shape}"
            )

            # Convert to `int` while torch support for `uint16` is not available
            img = img.int()
            has_instance = img >= divisor

            ids = (img % divisor) + 1
            sem = img // divisor
            sem[sem == void_id] = -1

            labels = PanopticTensor.from_parts(sem, ids)
        case PanopticFormat.KITTI:
            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            img = torchvision.io.decode_image(
                input, mode=torchvision.io.ImageReadMode.UNCHANGED
            )
            assert img.ndim == 3, f"Expected 3D tensor, got {img.ndim}D tensor"

            sem = img[:, :, 0]  # R-channel
            ids = torch.add(
                img[:, :, 1] * _BYTE_OFFSET,  # G channel
                img[:, :, 2],  # B channel
            )

            labels = PanopticTensor.from_parts(sem, ids)
        case PanopticFormat.VISTAS:
            divisor = 1000

            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            img = torchvision.io.decode_image(
                input, mode=torchvision.io.ImageReadMode.UNCHANGED
            )
            assert img.ndim == 2, f"Expected 2D tensor, got {img.ndim}D tensor"
            assert img.dtype == torch.int32, img.dtype

            labels = PanopticTensor.from_combined(img, divisor)
        case PanopticFormat.WILD_DASH:
            annotations = get_kwd(meta_kwds, "annotations", list[dict[str, T.Any]])

            divisor = int(1e8)
            void_id = 255

            if not isinstance(input, torch.Tensor):
                input = str(expath.locate(input))
            img = torchvision.io.decode_image(
                input, mode=torchvision.io.ImageReadMode.RGB
            )
            assert img.ndim == 3, f"Expected 3D tensor, got {img.ndim}D tensor"
            img = (
                img[:, :, 0].to(torch.long) * _BYTE_OFFSET * _BYTE_OFFSET
                + img[:, :, 1].to(torch.long) * _BYTE_OFFSET
                + img[:, :, 2].to(torch.long)
            )
            sem = torch.full_like(img, void_id, dtype=torch.long)
            for ann in annotations:
                id = ann["id"]
                category_id = ann["category_id"]
                mask = img == id
                sem[mask] = category_id

            ids = torch.full_like(img, 0, dtype=torch.long)  # TODO

            labels = PanopticTensor.from_parts(sem, ids)
            labels.translate_semantic_(
                translation=translations,
            )
        case _:
            msg = f"Could not read labels from {input!r} ({data_format=})"
            raise NotImplementedError(msg)
    assert labels is not None
    assert labels.ndim == 3, f"len({tuple(labels.shape)}) != 3"

    if len(meta_kwds) > 0:
        msg = f"Unexpected keyword arguments: {tuple(meta_kwds.keys())}"
        raise TypeError(msg)

    return labels


def save_panoptic(
    self: Tensor,
    path: expath.PathType,
    *,
    format: PanopticFormat | str | None = None,
    **kwargs,
) -> None:
    """Save the panoptic map to a file."""
    if self.ndim != 2:
        msg = f"Expected 2D tensor, got {self.ndim}D tensor ({self.shape})"
        raise ValueError(msg)

    path = expath.locate(path)

    # Check whether the format can be inferred from the filename
    if format is None:
        match path.suffix.lower():
            case ".pth", ".pt":
                format = PanopticFormat.TORCH
            case ".safetensors":
                format = PanopticFormat.SAFETENSORS
            case _:
                msg = f"Could not infer labels format from path: {path}"
                raise ValueError(msg)

    # Create the directory if it does not exist
    path.parent.mkdir(parents=True, exist_ok=True)

    # Save the labels
    match PanopticFormat(format):
        case PanopticFormat.SAFETENSORS:
            assert path.suffix.lower() == ".safetensors", (
                f"Expected SAFETENSORS file, got {path}"
            )
            safetensors.save_file({"data": torch.as_tensor(self)}, path)
        case PanopticFormat.TORCH:
            assert path.suffix.lower() in (
                ".pth",
                ".pt",
            ), f"Expected PT file, got {path}"
            torch.save(torch.as_tensor(self), path)
        case PanopticFormat.PNG_UINT16 | PanopticFormat.PNG_L16:
            assert path.suffix.lower() == ".png", f"Expected PNG file, got {path}"
            sem, ins = PanopticTensor.to_parts(self, as_tuple=True)

            void_id = kwargs.get("void_id", 64)
            max_ins = kwargs.get("divisor", 1000)

            assert not torch.any(sem == void_id), (
                f"Found void ID in semantic map, {void_id=}"
            )
            assert torch.all(ins < max_ins), f"Found instance ID >= {max_ins=}"

            pan = sem.where(sem >= 0, void_id) * max_ins + ins
            write_png_l16(path, pan)
        case PanopticFormat.CITYSCAPES:
            divisor = kwargs.get("divisor", 1000)
            void_id = kwargs.get("void_id", 255)
            img = torch.empty((*self.shape, 3), dtype=torch.uint8)
            img[:, :, 0] = self % _BYTE_OFFSET
            img[:, :, 1] = self // _BYTE_OFFSET
            img[:, :, 2] = self // _BYTE_OFFSET // _BYTE_OFFSET
            img = torch.where(self == void_id * divisor, 0, img)
            img = torch.where(self < divisor, img * divisor, img - 1)

            write_png_rgb(path, img)
        case PanopticFormat.CITYSCAPES_VPS:
            divisor = 1000
            void_id = 255
            sem, ids = PanopticTensor.to_parts(self, as_tuple=True)
            img = torch.where(ids > 0, ids - 1 + sem * divisor, sem)
            img = torch.where(img == void_id, 0, img)

            write_png_l16(path, img)
        case PanopticFormat.CITYSCAPES_DVPS:
            # https://github.com/joe-siyuan-qiao/ViP-DeepLab/tree/master/cityscapes-dvps
            max_ins = 255
            void_id = 32

            sem, ids = PanopticTensor.to_parts(self.detach().cpu(), as_tuple=True)
            assert (sem < void_id).all(), sem.unique().tolist()
            sem = sem.where(sem >= 0, void_id)
            assert (ids < max_ins).all(), ids.unique().tolist()

            img = torch.zeros((*self.shape[-2:], 3), dtype=torch.uint8)
            img[:, :, 0] = sem
            img[:, :, 1] = ids

            write_png_rgb(path, img)
        case PanopticFormat.KITTI:  # KITTI STEP
            img = torch.empty((*self.shape[-2:], 3), dtype=torch.uint8)

            sem, ids = PanopticTensor.to_parts(self, as_tuple=True)
            img[:, :, 0] = sem
            img[:, :, 1] = ids // _BYTE_OFFSET
            img[:, :, 2] = ids % _BYTE_OFFSET

            write_png_rgb(path, img)
        case PanopticFormat.VISTAS:
            divisor = 1000
            void_id = 255
            img = torch.empty((*self.shape, 3), dtype=torch.uint8)
            img[:, :, 0] = self % _BYTE_OFFSET
            img[:, :, 1] = self // _BYTE_OFFSET
            img[:, :, 2] = self // _BYTE_OFFSET // _BYTE_OFFSET

            write_png_rgb(path, img)
        case PanopticFormat.WILD_DASH:
            divisor = int(1e8)
            void_id = 255
            img = torch.empty((*self.shape, 3), dtype=torch.uint8)
            img[:, :, 0] = self // (_BYTE_OFFSET * _BYTE_OFFSET)
            img[:, :, 1] = (self // _BYTE_OFFSET) % _BYTE_OFFSET
            img[:, :, 2] = self % _BYTE_OFFSET
            img = torch.where(self == void_id * divisor, 0, img)

            write_png_rgb(path, img)
        case _:
            msg = f"Could not save labels to {path!r} ({format=})"
            raise NotImplementedError(msg)



"""
Canonical logger used throughout the project.

This module follows the same semantics as defined in the PyTorch logging.

See Also
--------
- `PyTorch Logging <https://pytorch.org/docs/stable/_modules/torch/_logging.html>`_
- `Design Document <https://docs.google.com/document/d/1ZRfTWKa8eaPq1AxaiHrq4ASTPouzzlPiuquSBEJYwS8/edit>`_
"""

from __future__ import annotations

import atexit
import collections
import enum as E
import functools
import io
import logging
import os
import sys
import time
import typing
import warnings
from collections import Counter
from typing import override

import expath
import pandas as pd
import tabulate
from termcolor import colored

import unipercept.utils.inspect as inspect_utils

__all__ = []

type Logger = logging.Logger


class LogLevel(E.StrEnum):
    DEBUG = E.auto()
    INFO = E.auto()
    WARNING = E.auto()
    ERROR = E.auto()
    CRITICAL = E.auto()


LOG_LEVELS: typing.Final[typing.OrderedDict[str, int]] = typing.OrderedDict(
    {
        LogLevel.DEBUG: logging.DEBUG,
        LogLevel.INFO: logging.INFO,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.ERROR: logging.ERROR,
        LogLevel.CRITICAL: logging.CRITICAL,
    }
)

type LogType = (
    typing.Literal[
        0,
        10,
        20,
        30,
        40,
        50,
        "debug",
        "info",
        "warning",
        "error",
        "critical",
    ]
    | LogLevel
)


@functools.cache
def _get_logging_level_map() -> collections.defaultdict[str, str]:
    """Get a mapping of (logger name) -> (logging level).

    By default, all loggers are set to the level of the root logger, which has
    a default value set by the environment variable ``UP_LOGS_LEVEL``, by default
    ``logging.INFO``.

    Individual modules' log levels can be overridden by setting the environment
    variable ``UP_LOGS``.
    The environment variable should be a comma-separated list of logger names and
    levels. We allow two formats for each entry:

        1. Explicitly pass the module name and log level as ``module name:log level``.
           Example: ``module1:info,module2:debug``.
        2. Prepend the module name with one or multiple `+` or `-` signs to respectively
           increasee verbosity. The adjustment is centered around the default log level
           as specified by the ``UP_LOGS_LEVEL`` environment variable (see above).
           Example: ``+module1,-module2,++module3``.

    Entries are parsed in the order above, so combinations are allowed, though
    this is not recommended for readability.
    """
    level_options = list(LOG_LEVELS.keys())
    level_base = os.getenv("UP_LOGS_LEVEL", "info").lower()
    result = collections.defaultdict(lambda: LOG_LEVELS[level_base])

    for spec in os.getenv("UP_LOGS", "").lower().split(","):
        spec = spec.strip()
        if not any(c in spec for c in "+-:"):
            continue
        spec, level = spec.split(":") if ":" in spec else (spec, level_base)
        if level not in level_options:
            msg = f"Unknown log level: {level}. Options are: {', '.join(level_options)}"
            raise ValueError(msg)
        module_name = spec.lstrip("+-")
        spec_adjust = spec[: len(spec) - len(module_name)]
        rel_amount = int(spec_adjust.count("+")) - int(spec_adjust.count("-"))
        cur_level = level_options.index(level)
        new_level = min(len(LOG_LEVELS), max(0, cur_level + rel_amount))

        result[module_name] = level_options[new_level]

    return result


def log_every(
    level: LogType, message: str, n: int = 1, *, name: str | None = None
) -> None:
    """Log once every `n` messages.

    Parameters
    ----------
    level : LogType
        The logging level.
    message : str
        The message to log.
    n : int
        Number of messages before a new message is written to the log.
    name : str
        Name of the logger to use. Defaults to the caller's module.
    """
    caller_module, key = inspect_utils.caller_identity()
    _log_counter[key] += 1
    if n == 1 or _log_counter[key] % n == 1:
        logging.getLogger(name or caller_module).log(_get_level(level), message)


def log_once(
    level: LogType, message: str, t: float = float("inf"), *, name: str | None = None
) -> None:
    """Log no more than once per `t` seconds (infinite by default).

    Parameters
    ----------
    level : LogType
        The logging level.
    message : str
        The message to log.
    t : float
        Amount of seconds before a new message is written to the log.
    name : str
        Name of the logger to use. Defaults to the caller's module.
    """
    caller_module, key = inspect_utils.caller_identity()
    last_logged = _log_timers.get(key)
    current_time = time.time()
    if last_logged is None or current_time - last_logged >= t:
        logging.getLogger(name or caller_module).log(_get_level(level), message)
        _log_timers[key] = current_time


class TableFormat(E.StrEnum):
    LONG = E.auto()
    WIDE = E.auto()
    AUTO = E.auto()


def create_table(
    mapping: typing.Mapping[typing.Any, typing.Any] | pd.DataFrame,
    format: TableFormat | typing.Literal["long", "wide", "auto"] = TableFormat.AUTO,
    *,
    style: str | None = None,
    max_depth: int | None = None,
    max_width: int | None = None,
    index: bool = False,
    _depth: int = 0,
    **kwargs,
) -> str:
    """
    Create a small table using the keys of small_dict as headers. This is only
    suitable for small dictionaries.

    Parameters
    ----------
    mapping
        Items to tabulate

    Returns
    -------
    The table as a string.
    """
    from os import getenv
    from pprint import pformat
    from shutil import get_terminal_size

    max_width = (
        int(getenv("UP_TABLE_MAX_WIDTH") or get_terminal_size((80, 20)).columns)
        if max_width is None
        else max_width
    )
    max_depth = max_depth or int(getenv("UP_TABLE_MAX_WIDTH") or 4)

    if style is None:
        style = getenv("UP_TABLE_STYLE", "rounded_outline")

    tabulate_kwargs: dict[str, typing.Any] = {
        "tablefmt": style,
        "floatfmt": ".4f",
        "stralign": "left",
        "numalign": "decimal",
        "missingval": "?",
        **kwargs,
        # "preserve_whitespace": True, # NOTE: not yet supported v0.9.0 (latest)
    }

    # Handling of various input types that need to be converted to a dictionary or
    # can be handled directly.
    if isinstance(mapping, pd.DataFrame):
        # In case of a Pandas DataFrame, we first try to convert it to a table by using
        # the built-in `to_markdown` method. Note that this calls `tabulate` under
        # the hood, so we do not actually get Markdown back, but rather the desired
        # table as a string following our own specification.
        try:
            result = mapping.to_markdown(index=index, **tabulate_kwargs)
        except Exception as e:
            warnings.warn(
                "Failed to convert DataFrame to table: " + str(e), stacklevel=2
            )
            result = None
        # Check whether the conversion was successful
        if isinstance(result, str):
            return result
        # Fallback to converting the dataframe to a dict in Pandas versions that do
        # not support passing Tabulate keyword-arguments
        mapping = mapping.to_dict(orient="list")
    elif (
        mapping_to_dict := getattr(mapping, "_asdict", None)
    ) is not None and callable(mapping_to_dict):
        mapping = mapping_to_dict()

    # Determine the format of the table
    if format == TableFormat.AUTO:
        AUTO_FORMAT_MAX_KEYS = 5
        AUTO_FORMAT_MAX_DEPTH = 0
        if len(mapping) <= AUTO_FORMAT_MAX_KEYS and _depth <= AUTO_FORMAT_MAX_DEPTH:
            format = TableFormat.WIDE
        else:
            format = TableFormat.LONG

    # Create the table in the specified format
    if format == TableFormat.WIDE:
        headers = list(mapping.keys())
        # Create a wide table, i.e. where each key is a header and the values are under
        # the corresponding header.
        # If the value is a non-sequence, it will be displayed as-is
        data = []
        for k in mapping:
            v = mapping[k]
            if isinstance(v, dict):
                v = create_table(
                    v,
                    format="auto",
                    style=style,
                    max_depth=max_depth,
                    _depth=_depth + 1,
                )
            if isinstance(v, str):
                v = v.split("\n")
            elif isinstance(v, typing.Sequence):
                # v = [pformat(v_item) for v_item in v]
                v = [str(v_item) for v_item in v]
            else:
                v = [v]
            data.append(v)

        col_lens = [len(v) for v in data]
        pad_to = max(col_lens) if len(col_lens) > 1 else 1

        for v in data:
            v.extend([""] * (pad_to - len(v)))

        # Transpose the data to make it wide
        data = list(zip(*data, strict=False))
    elif format == TableFormat.LONG:
        data = []
        for k in mapping:
            v = mapping[k]

            if isinstance(v, dict) and _depth < max_depth:
                for linenum, line in enumerate(
                    create_table(
                        v,
                        format="auto",
                        style=style,
                        max_depth=max_depth,
                        _depth=_depth + 1,
                    ).split("\n")
                ):
                    data.append((k if linenum == 0 else "", line))
                continue

            if not isinstance(v, str | int | float | bool) or v is None:
                try:
                    v = pformat(v)
                except Exception:
                    v = str(v)

            # Truncate long lines
            if isinstance(v, str):
                v_lines = v.split("\n")
                for v_i in range(len(v_lines)):
                    v_line = v_lines[v_i]
                    if len(v_line) > max_width:
                        v_line = v_line[: max_width - 3] + "..."

                    v_line = v_line.replace("\t", "  ")
                    # below is achieved via `preserve_whitespace` in tabulate_kwargs
                    # if (
                    #    v_line[0] == " "
                    # ):  # prevent tabulate from removing leading whitespace
                    #    v_line = "␣" + v_line[1:]
                    v_lines[v_i] = v_line

                # Put the first line in the same row as the key
                data.append((k, v_lines[0]))
                for v_line in v_lines[1:]:
                    data.append(("", v_line))
            else:
                data.append((k, v))
        # Create a long table, i.e. with a 'key' and 'value' header
        headers = ()  # ["Key", "Value"] if _depth == 0 else ()
    else:
        msg = f"Unknown table format: {format}"
        raise ValueError(msg)

    # Preserve whitespace in the table, this can be set with the `preserve_whitespace`
    # keyword argument in tabulate, but this is not yet supported in the current version
    # of tabulate (v0.9.0). We therefore set the global variable `PRESERVE_WHITESPACE`.
    old = tabulate.PRESERVE_WHITESPACE
    tabulate.PRESERVE_WHITESPACE = True
    tbl = tabulate.tabulate(data, headers=headers, showindex=index, **tabulate_kwargs)
    tabulate.PRESERVE_WHITESPACE = old

    return tbl


def get_logger(name: str | None = None, **kwargs: typing.Any) -> logging.Logger:
    """Get a logger instance, where the name is automatically set to the calling module name."""
    return _get_handler(
        _canonicalize_name(name or inspect_utils.calling_module_name(left=1)),
        **kwargs,
    )


@functools.cache
def _canonicalize_name(name: str) -> str:
    """Canonicalizes a logger name into a full module name without private submodules."""
    result = []
    for name in name.split(" "):
        if name.endswith(".py"):
            file = expath.locate(name)
            root = expath.locate(__file__).parent
            if file.is_relative_to(root):
                name = file.relative_to(root).with_suffix("").as_posix()
                name = name.replace("/", ".")
            else:
                msg = f"Cannot canonicalize name: {name}"
                raise ValueError(msg)

        try:
            name_parts = name.split(".")
            while name_parts[-1].startswith("_"):
                name_parts.pop()
            name = ".".join(name_parts)
        except IndexError:
            name = "<unknown>"

        result.append(name)
    return " ".join(result)


def _get_level(name_or_code: int | str) -> int:
    """Convert a log level name or code to a code."""
    match name_or_code:
        case int():
            return name_or_code
        case str():
            return LOG_LEVELS[name_or_code.lower()]
        case _:
            msg = f"Unknown log level: {name_or_code} ({type(name_or_code)})"
            raise TypeError(msg)


@functools.cache
def _get_handler(
    name: str,
    /,
    *,
    output=None,
    color=True,
    propagate: bool = False,
    stdout: bool = True,
    level: str | int | None = None,
):
    """
    Get a logger with a default verbose formatter, cached to ensure that the same logger handler is shared across
    all the sites that call this function.
    """
    from unipercept.state import get_process_index

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = propagate

    # Determine the full name and the short name
    root_name, *_sub_name = name.split(".", 1)
    module_name, *_artifact_name = name.split(" ", 1)

    # Translate the logging level
    if level is None:
        level = _get_logging_level_map()[module_name]
    level = _get_level(level)

    # Find the process index in distributed settings
    process_rank = max(0, get_process_index(local=False))

    # Logging to stdout
    if stdout and process_rank == 0:
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(level)
        ch.setFormatter(_get_formatter(root_name, color))

        logger.addHandler(ch)

    # Logging to file
    if output is not None:
        filename = expath.locate(output)
        if filename.suffix in (".txt", ".log"):
            filename = filename
        elif filename.suffix == "":
            filename = filename / "log.txt"
        else:
            msg = f"Unknown log file extension: {filename.suffix}"
            raise ValueError(msg)
        filename.with_suffix(f".{process_rank}.{filename.suffix}")
        filename.parent.mkdir(exist_ok=True, parents=True)

        fh = logging.StreamHandler(_get_stream(filename))
        fh.setLevel(level)
        fh.setFormatter(_get_formatter(full_name, short_name, False))

        logger.addHandler(fh)

    return logger


@functools.cache
def _get_formatter(root_name: str, color: bool = True) -> logging.Formatter:
    """Get a formatter instance, cached to ensure that the same formatter is shared across all the sites that call this."""
    datefmt = r"%Y-%m-%d %H:%M:%S"
    if color:
        prefix = "%(asctime)s %(name)s" + colored(":", attrs=("bold",))
        formatter = _ColorfulFormatter(
            prefix + " " + "%(message)s",
            datefmt=datefmt,
            root_name=root_name,
        )
    else:
        formatter = logging.Formatter(
            "[ %(asctime)s @ %(name)s ] (%(levelname)s) : %(message)s", datefmt=datefmt
        )

    return formatter


@functools.cache
def _get_stream(filename) -> io.IOBase:
    """Taken from the `detectron` implementation."""
    io = expath.open(filename, "a", buffering=1024**2)
    atexit.register(io.close)
    return io


class _ColorfulFormatter(logging.Formatter):
    """
    A logging formatter that supports colored output.
    Based on `detectron2.utils.logger._ColorfulFormatter`.
    """

    def __init__(self, *args, root_name: str, **kwargs) -> None:
        self._root_name = root_name + "."

        super().__init__(*args, **kwargs)

    @override
    def formatMessage(self, record):
        date, time = record.asctime.split(" ")
        record.asctime = (
            colored(date, color="cyan") + " " + colored(time, color="light_cyan")
        )

        match record.levelno:
            case logging.DEBUG:
                level_icon = "🐛"
            case logging.WARNING:
                level_icon = colored("⚠️", attrs=["blink"])
            case logging.ERROR:
                level_icon = colored("❌", attrs=["blink"])
            case logging.CRITICAL:
                level_icon = colored("😵", attrs=["blink"])
            case _:
                level_icon = "📝"

        root, *sub = record.name.split(".", 1)
        package_color = "yellow" if root == "unipercept" else "light_yellow"
        record.name = colored(root, attrs=["bold"], color=package_color)
        if sub and len(sub) > 0:
            record.name += colored("." + sub[0], color=package_color)
        record.name = f"{level_icon} {record.name}"  # {record.name:40s}"

        leader, sep, message = record.getMessage().partition(":")
        record.message = leader + sep + colored(message, attrs=["bold"])

        return super().formatMessage(record)


_log_counter = Counter()
_log_timers = {}


def log_first_n(
    level,
    message,
    n=1,
    *,
    name: str | None = None,
    key: tuple[str, ...] | str | typing.Literal["caller", "message"] = "caller",
) -> None:
    """
    Log only for the first n times.

    Parameters
    ----------
    lvl : int
        the logging level
    msg : str
        the message to log
    n : int
        the number of times to log
    name : str
        name of the logger to use. Will use the caller's module by default.
    key : str
        the key to use for counting. If "caller", will use the caller's module name.
        If "message", will use the message string.
    """
    if isinstance(key, str):
        key = (key,)
    assert len(key) > 0

    caller_module, caller_key = inspect_utils.caller_identity()
    hash_key = ()
    if "caller" in key:
        hash_key = hash_key + caller_key
    if "message" in key:
        hash_key = (*hash_key, message)

    _log_counter[hash_key] += 1
    if _log_counter[hash_key] <= n:
        logging.getLogger(name or caller_module).log(level, message)


########################
# Attribute forwarding #
########################


def __getattr__(name: str) -> typing.Any:
    """Forward all unknown attributes to the logging module."""
    if name == "logger" or name in (
        "debug",
        "info",
        "warning",
        "error",
        "critical",
        "exception",
    ):
        logger = get_logger(inspect_utils.calling_module_name(frames=0, strict=False))
        if name == "logger":
            return logger
        return getattr(logger, name)

    msg = f"Module {__name__} has no attribute {name}"
    raise AttributeError(msg)


if typing.TYPE_CHECKING:
    logger: typing.Final[logging.Logger] = logging.Logger("<caller>")
    debug = logger.debug
    info = logger.info
    warning = logger.warning
    error = logger.error
    critical = logger.critical
    exception = logger.exception



r"""
Flow tensors.
============

Provides support for optical and scene flow tensors.
"""

from torchvision.tv_tensors import Mask

from .registry import pixel_maps

__all__ = ["OpticalFlow", "SceneFlow"]


@pixel_maps.register()
class OpticalFlow(Mask):
    pass


@pixel_maps.register()
class SceneFlow(Mask):
    pass



"""
This package contains some classes from the Pytorch3D package. The combination
of Python and PyTorch versions required for the other packages in this project
did now allow the installation of the PyTorch3D from their official channels.

See: https://github.com/facebookresearch/pytorch3d
"""

from __future__ import annotations

import typing as T

import typing_extensions as TX

from . import render, transform
from .module import *



"""
An implementation of the DICE loss function for segmentation based on a padded tensor for separate background
and foreground classes.
"""

from __future__ import annotations

import torch
import typing_extensions as TX
from torch import nn

from .mixins import ScaledLossMixin, StableLossMixin


class WeightedStuffDiceLoss(StableLossMixin, ScaledLossMixin, nn.Module):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @TX.override
    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        y_num: int,
        y_mask: torch.Optional[torch.Tensor],
        index_mask: torch.Tensor,
    ) -> torch.Tensor:
        if y_num == 0:
            return x.mean() * 0.0

        _, _, h, w = y.shape

        x = x.reshape(-1, h, w)[index_mask, ...]
        y = y.reshape(-1, h, w)[index_mask, ...]

        x = torch.sigmoid(x)

        x = x.reshape(int(y_num), h * w)
        y = y.reshape(int(y_num), h * w)

        if y_mask is not None:
            y_mask = y_mask.reshape(-1, h, w)[index_mask, ...]
            y_mask = y_mask.reshape(int(y_num), h * w)

        return self._dice(x, y, y_mask, None)

    def _dice(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        y_valid: torch.Optional[torch.Tensor] = None,
        weights: torch.Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if y_valid is not None:
            invalid = torch.zeros_like(x)
            x = torch.where(y_valid, x, invalid)
            y = torch.where(y_valid, y, invalid)

        loss_part = (x**2).sum(dim=-1) + (y**2).sum(dim=-1)
        loss = 1.0 - 2.0 * (y * x).sum(dim=-1) / loss_part.clamp(self.eps)

        if weights is not None:
            loss = loss * weights

        return loss.sum() * self.scale


class WeightedThingDiceLoss(WeightedStuffDiceLoss):
    @TX.override
    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        y_num: int,
        y_mask: torch.Optional[torch.Tensor],
        index_mask: torch.Tensor,
        instance_num: int,
        weight_num: int,
        weights: torch.Tensor,
    ) -> torch.Tensor:
        if y_num == 0:
            return x.sigmoid().mean() * 0.0

        n, _, h, w = y.shape

        weights = weights.float()

        x = x.reshape(n, instance_num, weight_num, h, w)
        x = x.reshape(-1, weight_num, h, w)[index_mask, ...]

        y = y.unsqueeze(2).expand(n, instance_num, weight_num, h, w)
        y = y.reshape(-1, weight_num, h, w)[index_mask, ...]

        weights = weights.reshape(-1, weight_num)[index_mask, ...]
        weights = weights / weights.sum(dim=-1, keepdim=True).clamp(self.eps)

        x = torch.sigmoid(x)

        x = x.reshape(int(y_num), weight_num, h * w)
        y = y.reshape(int(y_num), weight_num, h * w)

        if y_mask is not None:
            y_mask = y_mask.unsqueeze(2).expand(n, instance_num, weight_num, h, w)
            y_mask = y_mask.reshape(-1, weight_num, h, w)[index_mask, ...]
            y_mask = y_mask.reshape(int(y_num), weight_num, h * w)

        return self._dice(x, y, y_mask, weights)


class WeightedThingFocalLoss(ScaledLossMixin, StableLossMixin, nn.Module):
    alpha: torch.jit.Final[float]
    gamma: torch.jit.Final[float]

    def __init__(self, alpha: float = 0.25, gamma: float = 2.0, **kwargs) -> None:
        super().__init__(**kwargs)

        self.alpha = alpha
        self.gamma = gamma

    @TX.override
    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        y_num: int,
        index_mask: torch.Tensor,
        instance_num: int,
        weights: torch.Tensor,
        weight_num: int,
    ) -> torch.Tensor:
        """
        Weighted version of Dice Loss used in PanopticFCN for multi-positive optimization.

        Adapted from: https://github.com/DdeGeus/PanopticFCN-IBS/blob/master/panopticfcn/loss.py

        Parameters
        ----------
        x
            prediction logits
        y
            segmentation target for Things or Stuff,
        gt_num
            ground truth number for Things or Stuff,
        index_mask
            positive index mask for Things or Stuff,
        instance_num
            instance number of Things or Stuff,
        weighted_val
            values of k positives,
        weighted_num
            number k for weighted loss,
        """
        # avoid Nan
        if y_num == 0:
            loss = x.sigmoid().mean() + self.eps
            return loss * y_num

        n, _, h, w = y.shape
        x = x.reshape(n, instance_num, weight_num, h, w)
        x = x.reshape(-1, weight_num, h, w)[index_mask, ...]
        y = y.unsqueeze(2).expand(n, instance_num, weight_num, h, w)
        y = y.reshape(-1, weight_num, h, w)[index_mask, ...]
        weights = weights.reshape(-1, weight_num)[index_mask, ...]
        weights = weights / torch.clamp(weights.sum(dim=-1, keepdim=True), min=self.eps)
        x = x.reshape(int(y_num), weight_num, h * w)
        y = y.reshape(int(y_num), weight_num, h * w)

        p = torch.sigmoid(x)
        ce_loss = nn.functional.binary_cross_entropy_with_logits(x, y, reduction="none")

        p_t = p * y + (1 - p) * (1 - y)
        loss = ce_loss * ((1 - p_t) ** self.gamma)

        if self.alpha >= 0:
            alpha_t = self.alpha * y + (1 - self.alpha) * (1 - y)
            loss = alpha_t * loss

        loss = loss.mean(dim=-1)
        loss = loss * weights

        return loss.sum() * self.scale



from __future__ import annotations

import types
import typing as T
from typing import override
from weakref import WeakValueDictionary

_O = T.TypeVar("_O", bound=T.Any)


class MissingValue:
    """A sentinel object that can be used as a value.

    Attributes
    ----------
    name : str
        The name of the sentinel value.

    """

    __sentinel_types__: T.ClassVar[WeakValueDictionary[str, T.Self]] = (
        WeakValueDictionary()
    )

    def __class_getitem__(cls, name: str) -> types.GenericAlias:
        return types.GenericAlias(cls, (name.upper()))

    def __new__(cls, name: str) -> T.Self:
        """
        Create a new sentinel value or return an existing one.

        Parameters
        ----------
        name : str
            The name of the sentinel value.

        Returns
        -------
        MissingValue
            The new or existing sentinel value.

        """
        name = name.upper()
        if name not in cls.__sentinel_types__:
            cls = super().__new__(cls)
            cls.__sentinel_types__[name] = cls
            # mcls.__sentinel_types__[name] = super().__new__(mcls, name, (), {})
        return cls.__sentinel_types__[name]

    def __init__(self, name: str) -> None:
        self.__name__: typing.Final[str] = name

    @override
    def __eq__(self, other: typing.Any) -> typing.TypeGuard[T.Self]:
        """
        Check if this sentinel value is equal to another object.

        Parameters
        ----------
        other : typing.Any
            The object to compare with this sentinel value.

        Returns
        -------
        bool
            True if the other object is a sentinel value with the same name,
            False otherwise.

        """
        return isinstance(other, MissingValue) and self.__name__ == other.__name__

    @override
    def __ne__(self, other: object) -> T.TypeGuard[T.Self]:
        """
        Check if this sentinel value is not equal to another object.

        Parameters
        ----------
        other : typing.Any
            The object to compare with this sentinel value.

        Returns
        -------
        bool
            True if the other object is not a sentinel value or has a different name,
            False otherwise.

        """
        return not (self == other)

    @override
    def __hash__(self) -> int:
        """
        Compute a hash value for this sentinel value.

        Returns
        -------
        int
            The hash value.

        """
        return hash(self.__name__)

    @override
    def __repr__(self) -> str:
        """
        Get a string representation of this sentinel value.

        Returns
        -------
        str
            The string representation.

        """
        return f"?{self.__name__}"

    def __contains__(self, other: typing.Any) -> typing.TypeGuard[T.Self]:
        return self.is_missing(other)

    def is_value(self, other: _O | T.Self) -> typing.TypeGuard[_O]:
        return not self.is_missing(other)

    def is_missing(self, other: typing.Any | T.Self) -> typing.TypeGuard[T.Self]:
        return other is self

    def guard_as(self, other: _O | T.Self, fn: typing.Callable[[T.Self], _O]) -> _O:
        if self.is_value(other):
            return other
        return fn(other)

    @classmethod
    def Error(cls):
        return TypeError(f"value is {cls.__name__}!")



from . import (
    box,
    check,
    descriptors,
    flops,
    formatter,
    frozendict,
    function,
    image,
    inspect,
    latency,
    mask,
    missing,
    pickle,
    python,
    signal,
    string,
    tensorclass,
    timestamp,
    ulid,
)

__all__ = [
    "box",
    "check",
    "descriptors",
    "flops",
    "formatter",
    "frozendict",
    "function",
    "image",
    "inspect",
    "latency",
    "mask",
    "missing",
    "pickle",
    "python",
    "signal",
    "string",
    "tensorclass",
    "timestamp",
    "ulid",
]



r"""
This module contains various implementations of computer vision algorithms,
conventions, and utilities.

Submodules
----------
- `filter`: Image filtering and convolution operations.
- `geometry`: Geometric transformations and conversions.
- `pointrend`: Point sampling and selection methods.
- `inpaint`: Image inpainting and completion methods.
"""

from . import filter, geometry, pointrend



"""
Utility to create a `tensordict`-like tensorclass using a superclass instead of
the provided decorator `@tensorclass`.

This is useful when you want typing to work properly, and is more explicit.
"""

from __future__ import annotations

import dataclasses as D
import types
import typing as T

import torch
import torch.distributed
import torch.types

# import torch.utils._pytree as pytree
from tensordict import TensorDictBase, is_tensor_collection, tensorclass

if T.TYPE_CHECKING:
    from tensordict.base import IndexType
    from tensordict.tensordict import NestedKey

type _CompatibleType = torch.Tensor


@T.dataclass_transform()
class _TensorclassMeta(type):
    """TensorclassMeta is a metaclass that wraps the `@tensorclass` decorator around the child."""

    def __new__(
        metacls, name: str, bases: tuple[type, ...], ns: dict[str, T.Any], **kwds
    ):
        # if len(bases) == 0:
        #     return super().__new__(cls, name, tuple(bases), ns, **kwds)

        bases = tuple(types.resolve_bases(bases))
        tc = super().__new__(metacls, name, bases, ns, **kwds)

        return tensorclass(tc)  # type: ignore


@T.dataclass_transform()
class Tensorclass(metaclass=_TensorclassMeta):
    """
    Tensorclass is a class that allows you to create a `tensordict`-like
    tensorclass using a superclass instead of the provided decorator `@tensorclass`.
    """

    def __post_init__(self):
        pass

    @classmethod
    def stack(cls, *others: T.Self) -> T.Self:
        """Stacks multiple tensorclasses together."""
        if len(others) == 0:
            msg = "Must provide at least one tensorclass to stack."
            raise ValueError(msg)

        if len(others) == 1:
            return others[0]

        td = T.cast(TensorDictBase, torch.stack(others))  # type: ignore
        return cls.from_tensordict(td)

    # Patch for `pin_memory` method
    def pin_memory(self, device=None, *, inplace=True) -> T.Self:
        """Pin memory for each Tensor field."""
        return self.apply(lambda x: x.pin_memory(device), inplace=inplace)

    # Type hints for methods and properties added by the @tensorclass decorator
    if T.TYPE_CHECKING:
        _: D.KW_ONLY
        batch_size: torch.Size | T.Sequence[int]
        device: torch.types.Device | str | None = None

        # Tensorclass-specific methods
        @classmethod
        def from_tensordict(
            cls,
            tensordict: TensorDictBase,
            non_tensordict: dict[str, T.Any] | None = None,
        ) -> T.Self: ...

        # Inherited methods
        @property
        def shape(self) -> torch.Size: ...

        @property
        def names(self) -> None: ...

        @names.setter
        def names(self, value) -> None: ...

        def refine_names(self, *names) -> None: ...

        def rename(self, *names, **rename_map) -> None: ...

        def rename_(self, *names, **rename_map) -> None: ...

        def size(self, dim: int | None = None) -> torch.Size | int: ...

        @property
        def requires_grad(self) -> bool: ...

        def ndimension(self) -> int: ...

        @property
        def ndim(self) -> int: ...

        def dim(self) -> int: ...

        def clear_device_(self) -> T.Self: ...

        def is_shared(self) -> bool: ...

        def state_dict(self) -> dict[str, T.Any]: ...

        def load_state_dict(self, state_dict: dict[str, T.Any]) -> T.Self: ...

        def is_memmap(self) -> bool: ...

        def numel(self) -> int: ...

        def send(
            self, dst: int, init_tag: int = 0, pseudo_rand: bool = False
        ) -> None: ...

        def recv(
            self, src: int, init_tag: int = 0, pseudo_rand: bool = False
        ) -> int: ...

        def isend(
            self, dst: int, init_tag: int = 0, pseudo_rand: bool = False
        ) -> int: ...

        def irecv(
            self,
        ) -> tuple[int, list[torch.Future]] | list[torch.Future] | None: ...

        def reduce(
            self,
            dst,
            op=torch.distributed.ReduceOp.SUM,
            async_op=False,
            return_premature=False,
        ) -> None: ...

        def pop(self) -> _CompatibleType: ...

        def apply_(self, fn: T.Callable) -> T.Self: ...

        def apply(self, fn: T.Callable, inplace: bool = False) -> T.Self: ...

        def as_tensor(self) -> None: ...

        def update(self) -> T.Self: ...

        def update_(self) -> T.Self: ...

        def update_at_(self) -> T.Self: ...

        def items(self) -> T.Iterator[tuple[str, _CompatibleType]]: ...

        def values(self) -> T.Iterator[_CompatibleType]: ...

        @property
        def sorted_keys(self) -> list[NestedKey]: ...

        def flatten(self, start_dim=0, end_dim=-1) -> T.Self: ...

        def unflatten(self, dim, unflattened_size) -> T.Self: ...

        def exclude(self, *keys: str, inplace: bool = False) -> T.Self: ...

        def copy_(self, tensordict: T.Self) -> T.Self: ...

        def copy_at_(self, tensordict: T.Self, idx: IndexType) -> T.Self: ...

        def get_at(self) -> _CompatibleType: ...

        def memmap_like(self, prefix: str | None = None) -> T.Self: ...

        def detach(self) -> T.Self: ...

        def to_h5(self) -> None: ...

        def to_tensordict(self) -> None: ...

        def zero_(self) -> T.Self: ...

        def unbind(self, dim: int) -> tuple[T.Self, ...]: ...

        def chunk(self, chunks: int, dim: int = 0) -> tuple[T.Self, ...]: ...

        def clone(self, recurse: bool = True) -> T.Self: ...

        def cuda(self, device: int = 0) -> T.Self: ...

        def masked_select(self, mask: torch.Tensor) -> T.Self: ...

        def to_dict(self) -> dict[str, T.Any]: ...

        def unsqueeze(self, dim: int) -> T.Self: ...

        def squeeze(self, dim: int | None = None) -> T.Self: ...

        def reshape(self) -> T.Self: ...

        def split(self, split_size: int | list[int], dim: int = 0) -> list[T.Self]: ...

        def gather(self) -> T.Self: ...

        def view(self) -> T.Self: ...

        def permute(self) -> T.Self: ...

        def all(self, dim: int | None = None) -> bool | T.Self: ...

        def any(self, dim: int | None = None) -> bool | T.Self: ...

        def get_sub_tensordict(self, idx: IndexType) -> T.Self: ...

        def flatten_keys(self) -> T.Self: ...

        def unflatten_keys(self) -> T.Self: ...

        def fill_(self, key: str, value: float | bool) -> T.Self: ...

        def empty(self) -> T.Self: ...

        def is_empty(self) -> bool: ...

        @property
        def is_locked(self) -> bool: ...

        @is_locked.setter
        def is_locked(self, value: bool) -> None: ...

        def lock_(self) -> T.Self: ...

        def unlock_(self) -> T.Self: ...

        @property
        def batch_dims(self) -> int: ...

        def expand(self, *shape: int) -> T.Self: ...

        def set(self) -> T.Self: ...

        def set_(self) -> T.Self: ...

        def del_(self, key: str) -> T.Self: ...

        def rename_key_(self) -> T.Self: ...

        def entry_class(self, key: NestedKey) -> type: ...

        def set_at_(self) -> T.Self: ...

        def get(self) -> T.Any: ...

        def share_memory_(self) -> T.Self: ...

        def detach_(self) -> T.Self: ...

        def memmap_(self) -> T.Self: ...

        @classmethod
        def load_memmap(cls, prefix: str) -> T.Self: ...

        def to(
            self, dest: torch.types.Device | torch.Size | type, **kwargs: T.Any
        ) -> T.Self: ...

        def masked_fill_(self, mask: torch.Tensor, value: float | bool) -> T.Self: ...

        def masked_fill(self, mask: torch.Tensor, value: float | bool) -> T.Self: ...

        def is_contiguous(self) -> bool: ...

        def contiguous(self) -> T.Self: ...

        def select(self) -> T.Self: ...

        def keys(self) -> T.Sequence[str]: ...


def is_tensordict_like(obj: T.Any) -> T.TypeGuard[TensorDictBase | Tensorclass]:
    """Check if the object is a `tensordict`-like object."""
    return isinstance(obj, (TensorDictBase, Tensorclass)) or is_tensor_collection(obj)
